import troposphere.cloudfront as clf

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj, change_obj_data)
from .route53 import R53_RecordSetCloudFront


class CFOriginAccessIdentity(clf.CloudFrontOriginAccessIdentity):
    def setup(self, comment):
        self.CloudFrontOriginAccessIdentityConfig = clf.CloudFrontOriginAccessIdentityConfig(
            Comment=comment,
        )

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

class CFDefaultCacheBehavior(clf.DefaultCacheBehavior):
    def setup(self, key):
        name = self.title
        if 'AllowedMethods' in key:
            self.AllowedMethods = get_endvalue(name + 'AllowedMethods')

        if 'CachedMethods' in key:
            self.CachedMethods = get_endvalue(name + 'CachedMethods')

        # If not defined default to True
        if 'Compress' in key:
            self.Compress = get_endvalue(name + 'Compress')
        else:
            self.Compress = True

        if 'DefaultTTL' in key:
            self.DefaultTTL = get_endvalue(name + 'DefaultTTL')

        self.ForwardedValues = clf.ForwardedValues()
        # If not defined default to True
        if 'QueryString' in key:
            self.ForwardedValues.QueryString=get_endvalue(name + 'QueryString')
        else:
            self.ForwardedValues.QueryString=True

        if 'CookiesForward' in key:
            self.ForwardedValues.Cookies = clf.Cookies(
                Forward=get_endvalue(name + 'CookiesForward')
            )
            if 'CookiesWhitelistedNames' in key:
                self.ForwardedValues.Cookies.WhitelistedNames = get_endvalue(name + 'CookiesWhitelistedNames', condition=True)
                # conditions
                add_obj(
                    get_condition(name + 'CookiesWhitelistedNames', 'equals', 'whitelist', name + 'CookiesForward')
                )

        # If not defined default to 'Host'
        if 'Headers' in key:
            self.ForwardedValues.Headers = get_endvalue(name + 'Headers')
        else:
            self.ForwardedValues.Headers = ['Host']

        if 'QueryStringCacheKeys' in key:
            self.ForwardedValues.QueryStringCacheKeys = get_endvalue(name + 'QueryStringCacheKeys', condition=True)
            # conditions
            add_obj(
                get_condition(name + 'QueryStringCacheKeys', 'equals', True, name + 'QueryString')
            )

        if 'TargetOriginId' in key:
            self.TargetOriginId = get_endvalue(name + 'TargetOriginId')
        else:
            self.TargetOriginId = If(
                'CloudFrontOriginAdHoc',
                Ref('RecordSetExternal'),
                Sub('${EnvRole}${RecordSetCloudFrontSuffix}.origin.' + cfg.HostedZoneNameEnv)
            )

        if 'MaxTTL' in key:
            self.MaxTTL = get_endvalue(name + 'MaxTTL')

        if 'MinTTL' in key:
            self.MinTTL = get_endvalue(name + 'MinTTL')

        if 'LambdaFunctionARN' in key:
            condname = name + 'LambdaFunctionARN'
            eventType = 'origin-request'
            if 'LambdaEventType' in key:
                eventType = key['LambdaEventType']
            # conditions
            add_obj(
                get_condition(condname, 'not_equals', 'None')
            )

            self.LambdaFunctionAssociations = [
                If(
                    condname,
                    clf.LambdaFunctionAssociation(
                        EventType=eventType,
                        LambdaFunctionARN=get_endvalue(condname)
                    ),
                    Ref('AWS::NoValue')
                )
            ]

        if 'ViewerProtocolPolicy' in key:
            self.ViewerProtocolPolicy = get_endvalue(name + 'ViewerProtocolPolicy')
        else:
            self.ViewerProtocolPolicy = 'redirect-to-https'


class CFCacheBehavior(clf.CacheBehavior, CFDefaultCacheBehavior):
    def setup(self, **kwargs):
        super(CFCacheBehavior, self).setup(**kwargs)
        name = self.title
        self.PathPattern = get_endvalue(name + 'PathPattern')


class CFDistributionConfig(clf.DistributionConfig):
    def setup(self):
        self.Enabled = 'True'
        DefaultCacheBehavior = CFDefaultCacheBehavior('CloudFrontCacheBehaviors0')
        DefaultCacheBehavior.setup(key=cfg.CloudFrontCacheBehaviors[0])
        self.DefaultCacheBehavior = DefaultCacheBehavior
        self.HttpVersion = get_endvalue('CloudFrontHttpVersion')
        self.Logging = If(
            'CloudFrontLogging',
            clf.Logging(
                Bucket=Sub(cfg.BucketLogs + '.s3.amazonaws.com'),
                Prefix=Sub('${EnvRole}.${AWS::StackName}/'),
            ),
            Ref('AWS::NoValue')
        )
        self.PriceClass = 'PriceClass_100'
        self.ViewerCertificate = clf.ViewerCertificate(
            AcmCertificateArn=If(
                'CloudFrontAcmCertificate',
                get_endvalue('GlobalCertificateArn'),
                Ref('AWS::NoValue')
            ),
            CloudFrontDefaultCertificate=If(
                'CloudFrontAcmCertificate',
                Ref('AWS::NoValue'),
                'True'
            ),
            SslSupportMethod=If(
                'CloudFrontAcmCertificate',
                'sni-only',
                Ref('AWS::NoValue')
            )
        )
        self.WebACLId = get_endvalue('CloudFrontWebACLId', condition=True)


class CFCustomOrigin(clf.CustomOriginConfig):
    def setup(self, key):
        name = self.title
        if 'HTTPSPort' in key:
            self.HTTPSPort = get_endvalue(name + 'HTTPSPort')
        if 'HTTPSort' in key:
            self.HTTPPort = get_endvalue(name + 'HTTPPort')
        self.OriginProtocolPolicy = get_endvalue(name + 'ProtocolPolicy')
        if 'SSLProtocols' in key:
            self.OriginSSLProtocols = get_endvalue(name + 'SSLProtocols')
        if 'ReadTimeout' in key:
            self.OriginReadTimeout = get_endvalue(name + 'ReadTimeout')
        if 'KeepaliveTimeout' in key:
            self.OriginKeepaliveTimeout = get_endvalue(name + 'KeepaliveTimeout')


class CFOriginCustomHeader(clf.OriginCustomHeader):
    def setup(self):
        name = self.title
        self.HeaderName = get_endvalue(name + 'Name')
        self.HeaderValue = get_endvalue(name + 'Value')


class CFOriginCLF(clf.Origin):
    def setup(self, index):
        name = self.title
        key = cfg.CloudFrontOrigins[index]

        CustomHeaders = []
        if 'Headers' in key:
            for n in key['Headers']:
                headername = '%sHeaders%s' % (name, n)
                CustomHeader = CFOriginCustomHeader(headername)
                CustomHeader.setup()

                CustomHeaders.append(CustomHeader)

        if key['Type'] == 'custom':
            CustomOrigin = CFCustomOrigin(name)
            CustomOrigin.setup(key=key)
            self.CustomOriginConfig = CustomOrigin
        else:
            self.S3OriginConfig = clf.S3OriginConfig()
            if 'OriginAccessIdentity' in key:
                self.S3OriginConfig.OriginAccessIdentity = get_subvalue('origin-access-identity/cloudfront/${1M}', name + 'OriginAccessIdentity')

        self.DomainName = get_endvalue(name + 'DomainName')
        self.OriginPath = get_endvalue(name + 'Path')
        self.Id = get_endvalue(name + 'Id')
        self.OriginCustomHeaders = CustomHeaders


class CFOriginEC2ECS(clf.Origin):
    def setup(self):
        cloudfrontorigincustomheaders = []

        self.DomainName = If(
            'CloudFrontOriginAdHoc',
            Ref('RecordSetExternal'),
            Sub('${EnvRole}${RecordSetCloudFrontSuffix}.origin.' + cfg.HostedZoneNameEnv)
        )
        self.Id = self.DomainName
        self.OriginPath = get_endvalue('CloudFrontOriginPath')

        try: custom_headers = cfg.CloudFrontOriginCustomHeaders
        except:
            pass
        else:
            for n in custom_headers:
                name = 'CloudFrontOriginCustomHeaders%s' % n
                cloudfrontorigincustomheaders.append(
                    If(
                        name,
                        clf.OriginCustomHeader(
                            HeaderName=get_endvalue(name + 'HeaderName'),
                            HeaderValue=get_endvalue(name + 'HeaderValue')
                        ),
                        Ref("AWS::NoValue")
                    )
                )
                # conditions
                add_obj(
                    get_condition(name, 'not_equals', '', name + 'HeaderName')
                )

        self.OriginCustomHeaders = cloudfrontorigincustomheaders
        self.CustomOriginConfig = clf.CustomOriginConfig(
            HTTPSPort=If(
                'CloudFrontOriginProtocolHTTPS',
                get_endvalue('ListenerLoadBalancerHttpsPort'),
                Ref('AWS::NoValue')
            ),
            HTTPPort=If(
                'CloudFrontOriginProtocolHTTP',
                get_endvalue('ListenerLoadBalancerHttpPort'),
                Ref('AWS::NoValue')
            ),
            OriginProtocolPolicy=get_endvalue('CloudFrontOriginProtocolPolicy')
        )


class CFOriginAGW(CFOriginEC2ECS):
    def setup(self):
        super(CFOriginAGW, self).setup()
        self.DomainName = Sub('${ApiGatewayRestApi}.execute-api.${AWS::Region}.amazonaws.com')
        self.Id = self.DomainName
        self.CustomOriginConfig = clf.CustomOriginConfig(
            HTTPSPort='443',
            OriginProtocolPolicy='https-only',
        )


class CFCustomErrorResponse(clf.CustomErrorResponse):
    def setup(self, key):
        name = self.title  # Ex. CustomErrorResponses1

        self.ErrorCode = get_endvalue(name + 'ErrorCode')
        self.ErrorCachingMinTTL = get_endvalue(name + 'ErrorCachingMinTTL')

        if 'ResponsePagePath' in key:
            self.ResponsePagePath = get_endvalue(name + 'ResponsePagePath')

        if 'ResponseCode' in key:
            self.ResponseCode = get_endvalue(name + 'ResponseCode')


def CFCustomErrors():
    CustomErrorResponses = []
    mapname = 'CloudFrontCustomErrorResponses'

    try: ErrorResponses = getattr(cfg, mapname)
    except:
        pass
    else:
        for n, v in ErrorResponses.items():
            name = '%s%s' % (mapname, n)
            CustomErrorResponse = CFCustomErrorResponse(name)
            CustomErrorResponse.setup(key=v)
            CustomErrorResponses.append(CustomErrorResponse)

    return CustomErrorResponses


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


# #################################
# ### START STACK INFRA CLASSES ###
# #################################

# S - CLOUDFRONT #
class CF_CloudFront(object):
    def __init__(self):
        # Resources
        CloudFrontDistribution = clf.Distribution('CloudFrontDistribution')

        DistributionConfig = CFDistributionConfig()
        DistributionConfig.setup()

        cachebehaviors = []
        sortedcachebehaviors = sorted(
            iter(cfg.CloudFrontCacheBehaviors.items()),
            key=lambda x_y: (
                (x_y[1]['Order']*1000) - len(x_y[1]['PathPattern']) if 'PathPattern' in x_y[1] else 0
            ) if 'Order' in x_y[1] else x_y[0]
        )
        itercachebehaviors = iter(sortedcachebehaviors)
        next(itercachebehaviors)
        for n, v in itercachebehaviors:
            if not v['PathPattern']:
                continue
            name = 'CloudFrontCacheBehaviors%s' % n

            cachebehavior = CFCacheBehavior(name)
            cachebehavior.setup(key=v)

            # Create and Use Condition only if PathPattern Value differ between envs
            if name + 'PathPattern' not in cfg.fixedvalues:
                # conditions
                add_obj(
                    get_condition(name, 'not_equals', 'None', name + 'PathPattern')
                )

                cachebehaviors.append(
                    If(
                        name,
                        cachebehavior,
                        Ref("AWS::NoValue")
                    )
                )
            else:
                cachebehaviors.append(cachebehavior)

        DistributionConfig.CacheBehaviors = cachebehaviors

        cloudfrontaliasextra = []
        for n in cfg.CloudFrontAliasExtra:
            name = 'CloudFrontAliasExtra%s' % n

            # parameters
            p_Alias = Parameter(name)
            p_Alias.Description = 'Alias extra - empty for default based on env/role'

            add_obj(p_Alias)
            
            cloudfrontaliasextra.append(get_endvalue(name, condition=True))

            # conditions
            add_obj(get_condition(name, 'not_equals', 'None'))

        CloudFrontDistribution.DistributionConfig = DistributionConfig
        CloudFrontDistribution.DistributionConfig.Aliases = cloudfrontaliasextra

        Origins = []
        try: cfg.CloudFrontOrigins
        except:
            pass
        else:
            for n, v in cfg.CloudFrontOrigins.items():
                name = 'CloudFrontOrigins%s' % n
                
                # parameters
                p_Origin = Parameter('CloudFrontOrigins' + n + 'DomainName')
                p_Origin.Description = 'Origin - empty for default based on env/role'
    
                p_OriginHTTPSPort = Parameter('CloudFrontOrigins' + n + 'HTTPSPort')
                p_OriginHTTPSPort.Description = 'Origin - empty for default based on env/role'
    
                add_obj([
                    p_Origin,
                    p_OriginHTTPSPort,
                ])
    
                Origin = CFOriginCLF(name)
                Origin.setup(index=n)
                Origins.append(Origin)

        CloudFrontDistribution.DistributionConfig.Origins = Origins
        CloudFrontDistribution.DistributionConfig.Comment = get_endvalue('CloudFrontComment')

        R_CloudFrontDistribution = CloudFrontDistribution

        add_obj([
            R_CloudFrontDistribution,
        ])

        self.CloudFrontDistribution = CloudFrontDistribution


class CF_CloudFrontInOtherService(CF_CloudFront):
    def __init__(self, key):
        super(CF_CloudFrontInOtherService, self).__init__()

        # Resources
        self.CloudFrontDistribution.DistributionConfig.Aliases[0:0] = [
            If(
                'RecordSetCloudFront',
                Sub('${EnvRole}${RecordSetCloudFrontSuffix}.cdn.' + cfg.HostedZoneNameEnv),
                Ref('RecordSetExternal')
            ),
            Sub('${EnvRole}${RecordSetCloudFrontSuffix}.' + cfg.HostedZoneNameEnv),
            If(
                'CloudFrontAliasZone',
                Sub('%s.%s' % (get_endvalue('CloudFrontAliasZone'), cfg.HostedZoneNameEnv)),
                Ref('AWS::NoValue')
            ),
        ]

        self.CloudFrontDistribution.Condition = 'CloudFrontDistribution'

        Origin = CFOriginEC2ECS()
        Origin.setup()
        self.CloudFrontDistribution.DistributionConfig.Origins.append(Origin)
        self.CloudFrontDistribution.DistributionConfig.Comment = Sub('${AWS::StackName}-${EnvRole}')

        try: cfg.CloudFrontCustomErrorResponses
        except:
            pass
        else:
            self.CloudFrontDistribution.DistributionConfig.CustomErrorResponses = CFCustomErrors()

        R53_RecordSetCloudFront()


class CF_CloudFrontEC2(CF_CloudFrontInOtherService):
    def __init__(self, key):
        super(CF_CloudFrontEC2, self).__init__(key)


class CF_CloudFrontECS(CF_CloudFrontEC2):
    def __init__(self, key):
        super(CF_CloudFrontECS, self).__init__(key)


class CF_CloudFrontAGW(CF_CloudFrontInOtherService):
    def __init__(self, key):
        super(CF_CloudFrontAGW, self).__init__(key)

        # To use the same code of ecs ec2 need to add "fake" condition, api-gateway use https only
        # Conditions
        add_obj([
            {'ListenerLoadBalancerHttpPort': Equals('True', 'False')},
            {'ListenerLoadBalancerHttpsPort': Equals('True', 'True')},
        ])

        Origin = CFOriginAGW()
        Origin.setup()
        self.CloudFrontDistribution.DistributionConfig.Origins = [Origin]

        # RecordSetExternal do not exist replace it with Ref('AWS::NoValue')
        change_obj_data(
            self.CloudFrontDistribution.DistributionConfig.Aliases,
            'RecordSetExternal',
            Ref('AWS::NoValue')
        )


class CF_CloudFrontCLF(CF_CloudFront):
    def __init__(self, key):
        super(CF_CloudFrontCLF, self).__init__()

        self.CloudFrontDistribution.DistributionConfig.CustomErrorResponses = CFCustomErrors()



import troposphere.waf as waf
import troposphere.wafregional as wafr

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class WAFIPSet(waf.IPSet):
    def setup(self, name):
        self.Condition = self.title  # Ex. WafIPSetAwsNat
        self.Name = name  # Ex. AwsNat


class WAFByteMatchSet(waf.ByteMatchSet):
    def setup(self, name):
        self.Condition = self.title
        self.Name = name


class WAFRule(waf.Rule):
    def setup(self, name):
        self.Condition = self.title  # Ex. WafRuleBlockIPsNotAllowed
        self.Name = name  # Ex. BlockIPsNotAllowed
        self.MetricName = self.Name


class WAFWebACL(waf.WebACL):
    def setup(self, name):
        self.Condition = self.title  # Ex. WafWebAclApplicationLoadBalancerExternal
        self.Name = name  # Ex. ApplicationLoadBalancerExternal
        self.MetricName = self.Name


class WAFGlobalIPSet(WAFIPSet):
    pass


class WAFGlobalByteMatchSet(WAFByteMatchSet):
    pass


class WAFGlobalRule(WAFRule):
    pass


class WAFGlobalWebAcl(WAFWebACL):
    pass


class WAFRegionalIPSet(wafr.IPSet, WAFIPSet):
    pass


class WAFRegionalByteMatchSet(wafr.ByteMatchSet, WAFByteMatchSet):
    pass


class WAFRegionalRule(wafr.Rule, WAFRule):
    pass


class WAFRegionalWebAcl(wafr.WebACL, WAFWebACL):
    pass

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

class WAFIPSetDescriptors(waf.IPSetDescriptors, wafr.IPSetDescriptors):
    def setup(self, name):
        self.Type = 'IPV4'
        self.Value=Select(int(self.title), get_endvalue(name + 'Ips'))


class WAFFieldToMatch(waf.FieldToMatch, wafr.FieldToMatch):
    def setup(self, key):
        name = self.title
        auto_get_props(self, key, del_prefix='FieldToMatch')


class WAFByteMatchTuples(waf.ByteMatchTuples, wafr.ByteMatchTuples):
    def setup(self, key):
        name = self.title  # Ex. WafByteMatchSets1Matches1
        auto_get_props(self, key)
        FieldToMatch = WAFFieldToMatch(self.title)
        FieldToMatch.setup(key=key)
        self.FieldToMatch = FieldToMatch


class WAFPredicates(waf.Predicates, wafr.Predicates):
    def setup(self, name, ptype, wtype):
        self.Negated = get_endvalue(name + 'Negated')
        if ptype == 'ByteMatch':
            self.DataId = Ref('Waf' + wtype + 'ByteMatchSet' + self.title)
            self.Type = 'ByteMatch'
        if ptype == 'IPMatch':
            self.DataId = Ref('Waf' + wtype + 'IPSet' + self.title)
            self.Type = 'IPMatch'
        if ptype == 'SizeConstraint':
            self.DataId = Ref('Waf' + wtype + 'SizeConstraintSet' + self.title)
            self.Type = 'SizeConstraint'
        if ptype == 'SqlInjectionMatch':
            self.DataId = Ref('Waf' + wtype + 'SqlInjectionMatchSet' + self.title)
            self.Type = 'SqlInjectionMatch'
        if ptype == 'XssMatch':
            self.DataId = Ref('Waf' + wtype + 'XssMatchSet' + self.title)
            self.Type = 'XssMatch'


class WAFAction(waf.Action, wafr.Action):
    pass


class WAFWebACLRule(waf.Rules, wafr.Rules):
    def setup(self, name, index, wtype):
        self.Action = WAFAction(
            Type=get_endvalue(name + 'Action')
        )
        self.Priority = index
        self.RuleId = Ref('Waf' + wtype + 'Rule' + self.title)


def WAF_condition(condname, mapname, wtype):
    return {condname: And(
        get_condition('', 'not_equals', 'None', mapname + 'Enabled'),
        Condition('Global') if wtype == 'Global' else Equals('1', '1'),
        Or( 
            get_condition('', 'equals', wtype, mapname + 'WafType'),
            get_condition('', 'equals', 'Common', mapname + 'WafType')
        )
    )}

# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


class WAF_IPSets(object):
    def __init__(self, key, wtype=''):
        for n, v in getattr(cfg, key).items():
            name = key.replace('Waf', wtype)  # Ex. RegionalIPSet
            resname = 'Waf' + name + n  # Ex. WafRegionalIPSetAwsNat 
            mapname = key + n  # Ex. WafIPSetAwsNat

            # conditions
            add_obj(WAF_condition(resname, mapname, wtype))

            # resources
            IPSetDescriptors = []
            for i, j in enumerate(v['Ips']):
                IPSetDescriptor = WAFIPSetDescriptors(str(i))
                IPSetDescriptor.setup(name=mapname)
                IPSetDescriptors.append(IPSetDescriptor)

            IPSet = globals()['WAF' + name](resname)
            IPSet.setup(name=n)
            IPSet.IPSetDescriptors = IPSetDescriptors

            add_obj(IPSet)


class WAF_ByteMatchSets(object):
    def __init__(self, key, wtype=''):
        for n, v in getattr(cfg, key).items():
            name = key.replace('Waf', wtype)  # Ex. RegionalByteMatchSet
            resname = 'Waf' + name + n
            mapname = key + n

            # conditions
            add_obj(WAF_condition(resname, mapname, wtype))

            # resources
            ByteMatchTuples = []
            for m, w in v['Matches'].items():
                matchname = '%sMatches%s' % (mapname, m)  # Ex. WafByteMatchSets1Matches1
                ByteMatchTuple = WAFByteMatchTuples(matchname)
                ByteMatchTuple.setup(key=w)

                ByteMatchTuples.append(ByteMatchTuple)

            ByteMatchSet = globals()['WAF' + name](resname)
            ByteMatchSet.setup(name=n)
            ByteMatchSet.ByteMatchTuples = ByteMatchTuples

            add_obj(ByteMatchSet)


class WAF_Rules(object):
    def __init__(self, key, wtype=''):
        for n, v in getattr(cfg, key).items():
            name = key.replace('Waf', wtype)  # Ex. RegionalIPSet
            resname = 'Waf' + name + n  # Ex. WafRegionalIPSetAwsNat 
            mapname = key + n  # Ex. WafIPSetAwsNat

            # conditions
            add_obj(WAF_condition(resname, mapname, wtype))

            # resources
            Predicates = []
            O_Predicates = []
            for m in v['Predicates']:
                predmapname = mapname + 'Predicates' + m  # Ex. WafByteMatchSetWafRules2Rules1
                predresname = m
                ptype = v['Type']
                Predicate = WAFPredicates(predresname)
                Predicate.setup(name=predmapname, ptype=ptype, wtype=wtype)

                Predicates.append(
                    If(
                        Predicate.DataId.data['Ref'],
                        Predicate,
                        Ref('AWS::NoValue')
                    )
                )

                O_Predicates.append(m)

            Rule = globals()['WAF' + name](resname)
            Rule.setup(name=n)
            Rule.Predicates = Predicates

            add_obj(Rule)

            # outputs
            O_Rule = Output(mapname + wtype)
            O_Rule.Condition = resname
            O_Rule.Value = ','.join(O_Predicates)

            add_obj(O_Rule)


class WAF_WebAcls(object):
    def __init__(self, key, wtype=''):
        for n, v in getattr(cfg, key).items():
            name = key.replace('Waf', wtype)  # Ex. RegionalIPSet
            resname = 'Waf' + name + n  # Ex. WafRegionalIPSetAwsNat 
            mapname = key + n  # Ex. WafIPSetAwsNat

            # conditions
            add_obj(WAF_condition(resname, mapname, wtype))

            # resources
            Rules = []
            O_Rules = []
            for w, m in enumerate(v['Rules'], start=1):
                rulemapname = '%sRules%s' % (mapname, m)
                ruleresname = m  # Ex. BlockIPsNotAllowed
                Rule = WAFWebACLRule(ruleresname)
                Rule.setup(name=rulemapname, index=w, wtype=wtype)

                Rules.append(Rule)

                O_Rules.append(m)

            WebACL = globals()['WAF' + name](resname)
            WebACL.setup(name=n)
            WebACL.Rules = Rules
            WebACL.DefaultAction = WAFAction(Type=v['DefaultAction'])

            add_obj(WebACL)

            # outputs
            O_WebACL = Output(mapname + wtype)
            O_WebACL.Condition = resname
            O_WebACL.Value = Sub('${%s} - %s' % (resname, ','.join(O_Rules)))

            add_obj(O_WebACL)

class WAF_GlobalByteMatchSets(object):
    def __init__(self, key):
        WAF_ByteMatchSets(key, wtype='Global')


class WAF_GlobalIPSets(object):
    def __init__(self, key):
        WAF_IPSets(key, wtype='Global')


class WAF_GlobalRules(WAF_Rules):
    def __init__(self, key):
        WAF_Rules(key, wtype='Global')


class WAF_GlobalWebAcls(WAF_WebAcls):
    def __init__(self, key):
        WAF_WebAcls(key, wtype='Global')


class WAF_RegionalByteMatchSets(object):
    def __init__(self, key):
        WAF_ByteMatchSets(key, wtype='Regional')


class WAF_RegionalIPSets(object):
    def __init__(self, key):
        WAF_IPSets(key, wtype='Regional')


class WAF_RegionalRules(WAF_Rules):
    def __init__(self, key):
        WAF_Rules(key, wtype='Regional')


class WAF_RegionalWebAcls(WAF_WebAcls):
    def __init__(self, key):
        WAF_WebAcls(key, wtype='Regional')



import troposphere.events as eve

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .lambdas import LambdaPermissionEvent

class EVERule(eve.Rule):
    def setup(self, key, name):
        auto_get_props(self, key)
        self.Name = Sub('${AWS::StackName}-${EnvRole}-' + 'Rule' + name)


class EVETarget(eve.Target):
    def setup(self, name, key):
        auto_get_props(self, key, mapname=name)

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class EVE_EventRules(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n  # Ex. EventsRuleElasticSearchSnapShot
            # parameters
            p_State = Parameter(resname + 'State')
            p_State.Description = 'Events Rule State - empty for default based on env/role'
            p_State.AllowedValues = ['', 'DISABLED', 'ENABLED']

            if 'ScheduleExpression' in v:
                p_ScheduleExpression = Parameter(resname + 'ScheduleExpression')
                p_ScheduleExpression.Description = 'Events Rule Schedule - empty for default based on env/role'

                add_obj(p_ScheduleExpression)

            add_obj([
                p_State,
            ])

            # resources
            Targets = []
            for m, w in v['Targets'].items():
                targetname = resname + 'Targets' + m
                Target = EVETarget('')
                Target.setup(name=targetname, key=w)
                Targets.append(Target)

                if m.startswith('Lambda'):
                    permname = m.replace('Lambda', 'LambdaPermission') + resname  # Ex. LambdaPermissionElasticSearchSnapShotEventsElasticsearchSnapShot
                    r_LambdaPermission = LambdaPermissionEvent(permname)
                    r_LambdaPermission.setup(key=w, source=resname)

                    add_obj(r_LambdaPermission)

            r_Rule = EVERule(resname)
            r_Rule.setup(key=v, name=n)
            r_Rule.Targets = Targets

            add_obj(r_Rule)

            # outputs
            o_State = Output(resname + 'State')
            o_State.Value = get_endvalue(resname + 'State')
   
            if 'ScheduleExpression' in v:
                o_ScheduleExpression = Output(resname + 'ScheduleExpression')
                o_ScheduleExpression.Value = get_endvalue(resname + 'ScheduleExpression')

                add_obj(o_ScheduleExpression)
    
            add_obj([
                o_State,
            ])



import troposphere.sns as sns

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .lambdas import LambdaPermissionSNS
from .sqs import SQSQueuePolicy

class SNSSubscription(sns.SubscriptionResource):
    def setup(self, key):
        auto_get_props(self, key)


class SNSTopic(sns.Topic):
    def setup(self, name):
        self.DisplayName = Sub('${AWS::StackName}.${EnvRole}-SNS%s' % name)

##
##

class SNS_Topics(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            resname = key + n  # Ex. SNSTopicASGNotification
            r_Topic = SNSTopic(resname)
            r_Topic.setup(name=n)
            if 'Condition' in v:
                r_Topic.Condition = v['Condition']

            add_obj(r_Topic)

            # outputs
            if 'Export' in v:
                o_Topic = Output(resname)
                o_Topic.Value = Ref(resname)
                o_Topic.Export = Export(resname)
                if 'Condition' in v:
                    o_Topic.Condition = v['Condition']

                add_obj(o_Topic)


class SNS_Subscriptions(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            if 'Enabled' in v and not v['Enabled']:
                continue
            resname = key + n  # Ex. SNSSubscriptionASGNotificationLambdaR53RecordInstanceId
            r_Subscription = SNSSubscription(resname)
            r_Subscription.setup(key=v)

            add_obj(r_Subscription)

            if v['Protocol'] == 'lambda':
                lambdaname = resname.split('Lambda')[1]
                topicname = resname.split('Lambda')[0].replace(key, 'SNS')
                permname = 'LambdaPermission' + lambdaname + topicname  # Ex. LambdaPermissionR53RecordInstanceIdSNSASGNotification 

                r_LambdaPermission = LambdaPermissionSNS(permname)
                r_LambdaPermission.setup(key=v)
                # Propagate condition, if present, to permission
                if hasattr(r_Subscription, 'Condition'):
                    r_LambdaPermission.Condition = r_Subscription.Condition

                add_obj(r_LambdaPermission)

            if v['Protocol'] == 'sqs':
                queuename = resname.split('SQS')[1]  # Ex. RabbitMQCluster
                topicname = resname.split('SQS')[0].replace(key, 'SNS')  # Ex. SNSASGNotification
                queuepolicyname = 'SQSQueuePolicy' + queuename + topicname  # Ex. SQSQueuePolicyRabbitMQClusterSNSASGNotification

                r_QueuePolicy = SQSQueuePolicy(queuepolicyname)
                r_QueuePolicy.setup(key=v)

                #add_obj(r_QueuePolicy)
                add_obj([r_QueuePolicy])



import troposphere.route53 as r53

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class R53RecordSet(r53.RecordSetType):
    def setup(self):
        pass


class R53RecordSetZoneExternal(R53RecordSet):
    def setup(self):
        super(R53RecordSetZoneExternal, self).setup()
        self.HostedZoneId = get_expvalue('HostedZoneIdEnv')
        self.Name = Sub('${AWS::StackName}.${EnvRole}.' + cfg.HostedZoneNameRegionEnv)  # Ex. prt-a-d.client-portal.eu-west-1.dev..


class R53RecordSetZoneInternal(R53RecordSet):
    def setup(self):
        super(R53RecordSetZoneInternal, self).setup()
        self.HostedZoneId = get_expvalue('HostedZoneIdPrivate')
        self.Name = Sub('${AWS::StackName}.${EnvRole}.' + cfg.HostedZoneNamePrivate)  # Ex. prt-a-d.client-portal.internal..


class R53RecordSetCloudFront(R53RecordSetZoneExternal):
    def setup(self):
        super(R53RecordSetCloudFront, self).setup()
        self.Condition = 'RecordSetCloudFront'
        self.AliasTarget = r53.AliasTarget(
            DNSName=GetAtt('CloudFrontDistribution', 'DomainName'),
            HostedZoneId=cfg.HostedZoneIdCF
        )
        self.Name = Sub('${EnvRole}${RecordSetCloudFrontSuffix}.cdn.' + cfg.HostedZoneNameEnv)  # Ex. client-portal.cdn.dev..
        self.Type = 'A'


class R53RecordSetLoadBalancer(R53RecordSet):
    def setup(self):
        super(R53RecordSetLoadBalancer, self).setup()
        self.AliasTarget = r53.AliasTarget(
            HostedZoneId=get_endvalue('HostedZoneIdLB')
        )
        self.Type = 'A'


class R53RecordSetEC2LoadBalancerExternal(R53RecordSetLoadBalancer, R53RecordSetZoneExternal):
    pass


class R53RecordSetEC2LoadBalancerInternal(R53RecordSetLoadBalancer, R53RecordSetZoneInternal):
    pass


class R53RecordSetECSLoadBalancer(R53RecordSetLoadBalancer):
    def setup(self, scheme):
        super(R53RecordSetECSLoadBalancer, self).setup()
        self.AliasTarget.DNSName = get_subvalue(
            'dualstack.${1E}',
            'LoadBalancerApplication%sDNS' % scheme,
            'LoadBalancerApplicationStack'
        ) 


class R53RecordSetECSLoadBalancerApplicationExternal(R53RecordSetECSLoadBalancer, R53RecordSetZoneExternal):
    pass


class R53RecordSetECSLoadBalancerApplicationInternal(R53RecordSetECSLoadBalancer, R53RecordSetZoneInternal):
    pass


class R53RecordSetEFS(R53RecordSet):
    def setup(self, efsname):
        super(R53RecordSetEFS, self).setup()
        condname = 'EFSFileSystem' + efsname  # Ex. EFSFileSystemWordPress
        self.Condition = condname
        self.HostedZoneId = Ref('HostedZonePrivate')
        self.Name = Sub('efs-%s.%s' % (efsname, cfg.HostedZoneNamePrivate))
        self.ResourceRecords = [
            Sub('${%s}.efs.${AWS::Region}.amazonaws.com' % condname)
        ]
        self.Type = 'CNAME'
        self.TTL = '300'


class R53RecordSetRDS(R53RecordSet):
    def setup(self):
        super(R53RecordSetRDS, self).setup()
        self.Type = 'CNAME'
        self.ResourceRecords = [GetAtt('DBInstance', 'Endpoint.Address')]
        self.TTL = '300'


class R53RecordSetRDSExternal(R53RecordSetRDS, R53RecordSetZoneExternal):
    pass


class R53RecordSetRDSInternal(R53RecordSetRDS, R53RecordSetZoneInternal):
    pass


class R53RecordSetCCH(R53RecordSet):
    def setup(self):
        super(R53RecordSetCCH, self).setup()
        self.Condition = 'CacheEnabled'
        self.Type = 'CNAME'
        if cfg.Engine == 'memcached':
            self.ResourceRecords = [GetAtt('CacheCluster', 'ConfigurationEndpoint.Address')]
        if cfg.Engine == 'redis':
            self.ResourceRecords = [GetAtt('CacheCluster', 'RedisEndpoint.Address')]
        self.TTL = '300'


class R53RecordSetCCHExternal(R53RecordSetCCH, R53RecordSetZoneExternal):
    pass


class R53RecordSetCCHInternal(R53RecordSetCCH, R53RecordSetZoneInternal):
    pass


class R53RecordSetNSServiceDiscovery(R53RecordSet):
    def setup(self):
        self.HostedZoneId = Ref('HostedZoneEnv')
        self.Name = Sub('find.' + cfg.HostedZoneNameEnv)
        self.ResourceRecords = GetAtt('PublicDnsNamespace', 'NameServers')
        self.Type = 'NS'
        self.TTL = '300'


# #################################
# ### START STACK INFRA CLASSES ###
# #################################


class R53_RecordSetCloudFront(object):
    def __init__(self):
       # Resources
        R_RecordSet = R53RecordSetCloudFront('RecordSetCloudFront')
        R_RecordSet.setup()

        add_obj([
            R_RecordSet,
        ])


class R53_RecordSetEC2LoadBalancer(object):
    def __init__(self):
        # Resources

        # RecordSet External
        if cfg.RecordSetExternal:
            R_External = R53RecordSetEC2LoadBalancerExternal('RecordSetExternal')
            R_External.setup()

            # LoadBalancerClassic
            if cfg.LoadBalancerClassicExternal:
                R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicExternal.DNSName}')
            elif cfg.LoadBalancerClassicInternal:
                R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicInternal.DNSName}')

            # LoadBalancerApplication
            if cfg.LoadBalancerApplicationExternal:
                R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationExternal.DNSName}')
            elif cfg.LoadBalancerApplicationInternal:
                R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationInternal.DNSName}')

            add_obj(R_External)

            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')

            add_obj(O_External)

        # RecordSet Internal
        if cfg.RecordSetInternal:
            R_Internal = R53RecordSetEC2LoadBalancerInternal('RecordSetInternal')
            R_Internal.setup()

            # LoadBalancerClassic
            if cfg.LoadBalancerClassicInternal:
                R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicInternal.DNSName}')
            elif cfg.LoadBalancerClassicExternal:
                R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicExternal.DNSName}')

            # LoadBalancerApplication
            if cfg.LoadBalancerApplicationInternal:
                R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationInternal.DNSName}')
            elif cfg.LoadBalancerApplicationExternal:
                R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationExternal.DNSName}')

            add_obj(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')

            add_obj(O_Internal)


class R53_RecordSetECSLoadBalancer(object):
    def __init__(self):
        # Resources
        if cfg.RecordSetExternal:
            R_External = R53RecordSetECSLoadBalancerApplicationExternal('RecordSetExternal')

            if cfg.LoadBalancerApplicationExternal:
                R_External.setup(scheme='External')
            else:
                R_External.setup(scheme='Internal')

            add_obj(R_External)
            
            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')

            add_obj(O_External)

        if cfg.RecordSetInternal:
            R_Internal = R53RecordSetECSLoadBalancerApplicationInternal('RecordSetInternal')

            if cfg.LoadBalancerApplicationInternal:
                R_Internal.setup(scheme='Internal')
            else:
                R_Internal.setup(scheme='External')

            add_obj(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')

            add_obj(O_Internal)


class R53_RecordSetRDS(object):
    def __init__(self):
        # Resources
        if cfg.RecordSetExternal:
            R_External = R53RecordSetRDSExternal('RecordSetExternal')
            R_External.setup()
            add_obj(R_External)

            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')

            add_obj(O_External)

        if cfg.RecordSetInternal:
            R_Internal = R53RecordSetRDSInternal('RecordSetInternal')
            R_Internal.setup()
            add_obj(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')

            add_obj(O_Internal)


class R53_RecordSetCCH(object):
    def __init__(self):
        # Resources
        if cfg.RecordSetExternal:
            R_External = R53RecordSetCCHExternal('RecordSetExternal')
            R_External.setup()
            add_obj(R_External)

            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')
            O_External.Condition = 'CacheEnabled'

            add_obj(O_External)

        if cfg.RecordSetInternal:
            R_Internal = R53RecordSetCCHInternal('RecordSetInternal')
            R_Internal.setup()
            add_obj(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')
            O_Internal.Condition = 'CacheEnabled'

            add_obj(O_Internal)


class R53_HostedZones(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            mapname = key + n
            resname = v['ResourceName']
            output_zonename = resname.replace('HostedZone', 'HostedZoneName')
            output_zoneidname = resname.replace('HostedZone', 'HostedZoneId')
            # parameters
            if n.startswith('Public'):
                p_HostedZone = Parameter(mapname + 'Enabled')
                p_HostedZone.Description = 'Create Public %s - can be created in only one Region - empty for default based on env/role' % resname

                p_HostedZoneId = Parameter(mapname + 'Id')
                p_HostedZoneId.Description = 'Id of Public %s - required in all Regions where HostedZonePublicEnv is not created - empty for default based on env/role' % resname

                add_obj([
                    p_HostedZone,
                    p_HostedZoneId,
                ])

                # conditions
                add_obj([
                    get_condition(resname, 'not_equals', 'None', mapname + 'Enabled')
                ])

            # resources
            r_HostedZone = r53.HostedZone(v['ResourceName'])
            auto_get_props(r_HostedZone, v, recurse=True, mapname=mapname)
            if n.startswith('Public'):
                r_HostedZone.Condition = resname

            add_obj(r_HostedZone)

            # outputs
            o_HostedZoneName = Output(output_zonename)
            #o_HostedZoneName.Value = Sub(cfg.HostedZoneNamePrivate)
            o_HostedZoneName.Value = get_endvalue(mapname + 'Name')
                
            o_HostedZoneId = Output(output_zoneidname)
            o_HostedZoneId.Value = If(resname, Ref(resname), get_endvalue(mapname + 'Id')) if n.startswith('Public') else Ref(resname)
            o_HostedZoneId.Export = Export(output_zoneidname)

            add_obj([
                o_HostedZoneName,
                o_HostedZoneId,
            ])



import troposphere.ecs as ecs

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .securitygroup import (SecurityGroupEcsService, SecurityGroupRuleEcsService,
    SG_SecurityGroupsECS)

class ECSService(ecs.Service):
    def setup(self, scheme):
        self.Cluster = get_expvalue('Cluster', 'ClusterStack')

        # DAEMON MODE DO NOT SUPPORT DESIRED OR PLACEMENT STRATEGIES, IS SIMPLY ONE TASK FOR CONTAINER INSTANCE
        if cfg.SchedulingStrategy == 'REPLICA':
            self.DesiredCount = get_endvalue('CapacityDesired')
            self.DeploymentConfiguration = ecs.DeploymentConfiguration(
                MaximumPercent = get_endvalue('DeploymentConfigurationMaximumPercent'),
                MinimumHealthyPercent = get_endvalue('DeploymentConfigurationMinimumHealthyPercent'),
            )
            self.PlacementStrategies = If(
                'LaunchTypeFarGate',
                Ref('AWS::NoValue'),
                [
                    #ecs.PlacementStrategy(
                    #    Type='spread',
                    #    Field='attribute:ecs.availability-zone'
                    #),
                    ecs.PlacementStrategy(
                        Type='spread',
                        Field='instanceId'
                    )
                ]
            )
        elif cfg.SchedulingStrategy == 'DAEMON':
            self.SchedulingStrategy = 'DAEMON'

        self.LaunchType = get_endvalue('LaunchType')

        if cfg.HealthCheckGracePeriodSeconds != 0:
            self.HealthCheckGracePeriodSeconds = get_endvalue('HealthCheckGracePeriodSeconds')

        #self.PlatformVersion = 'LATEST'

        if cfg.LoadBalancerApplication:
            self.LoadBalancers = []
            self.Role = If(
                'NetworkModeAwsVpc',
                Ref('AWS::NoValue'),
                get_expvalue('RoleECSService')
            )
            # When creating a service that specifies multiple target groups, the Amazon ECS service-linked role must be created.
            # The role is created by omitting the Role property in AWS CloudFormation
            if cfg.LoadBalancerApplicationExternal and cfg.LoadBalancerApplicationInternal:
                self.Role = Ref('AWS::NoValue') 

        self.TaskDefinition = Ref('TaskDefinition')


class ECSTaskDefinition(ecs.TaskDefinition):
    def setup(self):
        self.ExecutionRoleArn = If(
            'LaunchTypeFarGate',
            get_expvalue('RoleECSTaskExecution'),
            Ref('AWS::NoValue')
        )
        self.TaskRoleArn = Ref('RoleTask')

        self.Cpu = If(
            'CpuTask',
            get_endvalue('Cpu'),
            Ref('AWS::NoValue')
        )

        self.Memory = If(
            'LaunchTypeFarGate',
            get_endvalue('Memory'),
            Ref('AWS::NoValue')
        )

        self.NetworkMode = If(
            'NetworkModeAwsVpc',
            'awsvpc',
            Ref('AWS::NoValue')
        )

        self.RequiresCompatibilities = If(
            'LaunchTypeFarGate',
            ['EC2', 'FARGATE'],
            ['EC2']
        )
        

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

class ECSEnvironment(ecs.Environment):
    def setup(self, key):
        name = self.title
        valuename = name + 'Value'
        self.Name = get_endvalue(name + 'Name')
        self.Value = get_endvalue(valuename)


class ECSMountPoint(ecs.MountPoint):
    def __init__(self, title, key, **kwargs):
        super(ECSMountPoint, self).__init__(title, **kwargs)
        self.ReadOnly = False
        self.SourceVolume = self.title
        self.ContainerPath = key['ContainerPath']


class ECSLoadBalancer(ecs.LoadBalancer):
    def __init__(self, title, scheme, **kwargs):
        super(ECSLoadBalancer, self).__init__(title, **kwargs)
        self.ContainerName = Ref('EnvRole')
        self.ContainerPort = get_endvalue('ContainerDefinitions1ContainerPort')
        self.TargetGroupArn = Ref('TargetGroup' + scheme)


class ECSContainerDefinition(ecs.ContainerDefinition):
    def setup(self, key, index):
        name = self.title  # Ex. ContainerDefinitions1
        auto_get_props(self, key)

        self.Essential = True

        if len(cfg.ContainerDefinitions) == 1:
            self.Cpu = If(
                'CpuTask',
                get_endvalue('Cpu'),
                get_endvalue(name + 'Cpu')
            )
            self.Memory = If(
                'LaunchTypeFarGate',
                get_endvalue('Memory'),
                get_endvalue(name + 'Memory')
            )

        if 'RepoName' in key:
            self.Image=get_subvalue(
                '${1M}.dkr.ecr.${AWS::Region}.amazonaws.com/${2M}:${EnvApp%sVersion}' % index,
                ['EcrAccount', name + 'RepoName']
            )
        # use the same EnvApp version for all containers
        elif cfg.RepoName != 'None':
            self.Image=get_subvalue(
                '${1M}.dkr.ecr.${AWS::Region}.amazonaws.com/${2M}:${EnvApp1Version}',
                ['EcrAccount', 'RepoName']
            )
        elif cfg.Image != 'None':
            self.Image=get_endvalue('Image')

        self.LogConfiguration=If(
            'LogConfiguration',
            ecs.LogConfiguration(
                LogDriver=get_endvalue('LogDriver'),
                Options={
                    #'awslogs-group': get_endvalue('AwsLogsGroup'),
                    #'awslogs-create-group': True,
                    'awslogs-group': Ref('LogsLogGroup'),
                    'awslogs-region': Ref('AWS::Region'),
                    'awslogs-stream-prefix': Ref('AWS::StackName')
                }
            ),
            Ref('AWS::NoValue')
        )

        if 'MountPoints' in key:
            self.MountPoints = [
                ECSMountPoint(n, key=k) for n, k in key['MountPoints'].items()
            ]

        if 'Name' in key:
            self.Name = get_subvalue('${EnvRole}-${1M}', name + 'Name')
        else:
            self.Name = Ref('EnvRole')

        if 'ContainerPort' in key:
            PortMapping = ecs.PortMapping()
            auto_get_props(PortMapping, key, mapname=self.title)
            if 'HostPort' not in key:
                PortMapping.HostPort = If(
                    'NetworkModeAwsVpc',
                    get_endvalue(name + 'ContainerPort'),
                    0
                )
            self.PortMappings = [PortMapping]


class ECSVolume(ecs.Volume):
    def __init__(self, title, key, **kwargs):
        super(ECSVolume, self).__init__(title, **kwargs)
        self.Name = self.title
        self.Host = ecs.Host(
            SourcePath=key['SourcePath']
        )


class ECSAwsvpcConfiguration(ecs.AwsvpcConfiguration):
    def setup(self):
        self.SecurityGroups = [
            GetAtt('SecurityGroupEcsService', 'GroupId'),
        ]
        self.Subnets = Split(',', get_expvalue('SubnetsPrivate'))


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

class ECS_TaskDefinition(object):
    def __init__(self, key):
        Environments_Base = [
            ecs.Environment(
                Name='EnvRegion',
                Value=Ref('AWS::Region')
            ),
            ecs.Environment(
                Name='Env',
                Value=Ref('Env')
            ),
            ecs.Environment(
                Name='EnvAbbr',
                Value=Ref('EnvShort')
            ),
            ecs.Environment(
                Name='EnvBrand',
                Value=cfg.BrandDomain
            ),
            ecs.Environment(
                Name='EnvRole',
                Value=Ref('EnvRole')
            ),
            ecs.Environment(
                Name='EnvStackName',
                Value=Ref('AWS::StackName')
            ),
            ecs.Environment(
                Name='EnvClusterStackName',
                Value=get_endvalue('ClusterStack')
            ),
        ]

        Containers = []
        for n, v in cfg.ContainerDefinitions.items():
            Environments = []
            MountPoints = []

            name = 'ContainerDefinitions%s' % n  # Ex. ContainerDefinitions1

            # parameters
            # if ContainerDefinitions have RepoName use different EnvApp version
            if n == 1 or 'RepoName' in v:  
                nameenvapp = 'EnvApp%sVersion' % n  # Ex. EnvApp1Version

                EnvApp = Parameter(nameenvapp)
                EnvApp.Description = nameenvapp
                EnvApp.AllowedPattern = '^[a-zA-Z0-9-_.]*$'
                EnvApp.Default = '1'

                add_obj(EnvApp)

                # outputs
                o_EnvAppOut = Output(nameenvapp)
                o_EnvAppOut.Value = Ref(nameenvapp)

                # and use different output for RepoName
                if cfg.RepoName != 'None':
                    if 'RepoName' in v:
                        o_Repo = Output(name + 'RepoName')
                        o_Repo.Value = get_endvalue(name + 'RepoName')
                    else:
                        o_Repo = Output('RepoName')
                        o_Repo.Value = get_endvalue('RepoName')

                    add_obj(o_Repo)

                add_obj([
                    o_EnvAppOut,
                ])

            # outputs
            EnvValue_Out_String = []
            EnvValue_Out_Map = {}
            if 'Envs' in v:
                for m, w in v['Envs'].items():
                    envname = '%sEnvs%s' % (name, m)
                    # parameters
                    EnvValue = Parameter(envname + 'Value')
                    EnvValue.Description = '%s - empty for default based on env/role' % w['Name']

                    # If key NoParam is present skip adding Parameters (usefull as they have a limited max number)
                    if 'NoParam' not in w:
                        add_obj(EnvValue)

                    Environment = ECSEnvironment(envname)
                    Environment.setup(key=w)
                    Environments.append(Environment)

                    # outputs
                    EnvValue_Out_String.append('%s=${%s}' % (w['Name'], w['Name']))
                    EnvValue_Out_Map.update({
                        w['Name']: Environment.Value
                    })

            o_EnvValueOut = Output(name + 'Envs')
            o_EnvValueOut.Value = Sub(','.join(EnvValue_Out_String), **EnvValue_Out_Map)

            add_obj(o_EnvValueOut)

            Environments.extend(Environments_Base)

            # parameters
            if 'Cpu' in v:
                p_Cpu = Parameter(name + 'Cpu')
                p_Cpu.Description = 'Cpu Share for containers - empty for default based on env/role'

                add_obj(p_Cpu)

            if 'Memory' in v:
                p_Memory = Parameter(name + 'Memory')
                p_Memory.Description = 'Memory hard limit for containers - empty for default based on env/role'

                add_obj(p_Memory)

            if 'MemoryReservation' in v:
                p_MemoryReservation = Parameter(name + 'MemoryReservation')
                p_MemoryReservation.Description = 'Memory soft limit for containers - empty for default based on env/role'

                add_obj(p_MemoryReservation)

            if 'Command' in v:
                p_Command = Parameter(name + 'Command')
                p_Command.Description = 'Command to execute'
                p_Command.Type = 'CommaDelimitedList'

                add_obj(p_Command)

            Container = ECSContainerDefinition(name)
            Container.setup(key=v, index=n)
            Container.Environment = Environments
            # Trick to force reload of Service using parameter
            Container.DockerLabels = If(
                'DockerLabelLastUpdateOverride',
                {'LastUpdate': Ref('DockerLabelLastUpdate')},
                Ref('AWS::NoValue'),
            )
            
            Containers.append(Container)

            # outputs
            Constraints_Out_String = []
            Constraints_Out_Map = {}

            if 'Cpu' in v:
                Constraints_Out_String.append('Cpu:${Cpu}')
                Constraints_Out_Map.update({
                    'Cpu': Container.Cpu
                })

            if 'Memory' in v:
                Constraints_Out_String.append('Memory:${Memory}')
                Constraints_Out_Map.update({
                    'Memory': Container.Memory
                })

            if 'MemoryReservation' in v:
                Constraints_Out_String.append('MemoryReservation:${MemoryReservation}')
                Constraints_Out_Map.update({
                    'MemoryReservation': Container.MemoryReservation
                })

            if Constraints_Out_String:
                o_Constraints = Output(name + 'Constraints')
                o_Constraints.Value = Sub(','.join(Constraints_Out_String), **Constraints_Out_Map)

                add_obj(o_Constraints)

            if 'Command' in v:
                o_Command = Output(name + 'Command')
                o_Command.Value = Join(',', get_endvalue(name + 'Command'))

                add_obj(o_Command)

        # Resources
        R_TaskDefinition = ECSTaskDefinition('TaskDefinition')
        R_TaskDefinition.setup()
        R_TaskDefinition.ContainerDefinitions = Containers

        if cfg.Volumes:
            Volumes = []
            for n, v in cfg.Volumes.items():
                Volume = ECSVolume(n, key=v)
                Volumes.append(Volume)

            R_TaskDefinition.Volumes = Volumes

        add_obj([
            R_TaskDefinition,
        ])


class ECS_Service(object):
    def __init__(self, key):
        # Resources
        R_SG = SecurityGroupEcsService('SecurityGroupEcsService')
        R_SG.setup()

        R_Service = ECSService('Service')
        R_Service.setup(scheme='')

        if cfg.LoadBalancerApplicationExternal:
            R_Service.LoadBalancers.append(ECSLoadBalancer('', scheme='External'))

            SGRule = SecurityGroupRuleEcsService()
            SGRule.setup(scheme='External')
            R_SG.SecurityGroupIngress.append(SGRule)

        if cfg.LoadBalancerApplicationInternal:
            R_Service.LoadBalancers.append(ECSLoadBalancer('', scheme='Internal'))

            SGRule = SecurityGroupRuleEcsService()
            SGRule.setup(scheme='Internal')
            R_SG.SecurityGroupIngress.append(SGRule)

        SecurityGroups = SG_SecurityGroupsECS().SecurityGroups
        NetworkConfiguration = ecs.NetworkConfiguration()
        NetworkConfiguration.AwsvpcConfiguration = ECSAwsvpcConfiguration()
        NetworkConfiguration.AwsvpcConfiguration.setup()
        NetworkConfiguration.AwsvpcConfiguration.SecurityGroups.extend(SecurityGroups)

        R_Service.NetworkConfiguration = If(
            'NetworkModeAwsVpc',
            NetworkConfiguration,
            Ref('AWS::NoValue')
        )

        add_obj([
            R_Service,
            R_SG,
        ])


class ECS_Cluster(object):
    def __init__(self, key):
        # Resources
        R_Cluster = ecs.Cluster('Cluster')

        add_obj([
            R_Cluster,
        ])



import troposphere.codedeploy as cdd

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class CDEc2TagFilters(cdd.Ec2TagFilters):
    def __init__(self, title, **kwargs):
        super(CDEc2TagFilters, self).__init__(title, **kwargs)
        self.Key = 'EnvStackName'
        self.Type = 'KEY_AND_VALUE'
        self.Value = Ref('AWS::StackName')


class CDDeploymentGroup(cdd.DeploymentGroup):
    def setup(self, index):
        appreponame = 'Apps%sRepoName' % index
        appenvname = 'EnvApp%sVersion' % index
        self.Condition = 'DeploymentGroup'
        self.ApplicationName = get_endvalue(appreponame)
        # Uncomment for old behaviour where codedeploy prepare at boot autoscalinggroup's instances (do not work with autospot)
        #self.AutoScalingGroups = If(
        #    'DeployRevision',
        #    [Ref('AutoScalingGroup')],
        #    Ref('AWS::NoValue')
        #)
        # AUTOSPOT - Use Ec2TagFilters to let it work with instances launched by autospot
        self.Ec2TagFilters = If(
            'DeployRevision',
            [CDEc2TagFilters('')],
            Ref('AWS::NoValue')
        )
        self.Deployment = If(
            'DeployRevision',
            cdd.Deployment(
                Revision=cdd.Revision(
                    RevisionType='S3',
                    S3Location=cdd.S3Location(
                        Bucket=Sub(cfg.BucketAppRepository),
                        BundleType='tgz',
                        Key=get_subvalue('${1M}/${1M}-${%s}.tar.gz' % appenvname, appreponame)
                    )   
                )   
            ),  
            Ref('AWS::NoValue')
        )   
        self.DeploymentGroupName = Sub('${AWS::StackName}.${EnvRole}')
        self.ServiceRoleArn = get_expvalue('RoleCodeDeploy', '')

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class CD_DeploymentGroup(object):
    def __init__(self):
        # Conditions
        add_obj([
            {'DeploymentGroup': And(
                Condition('Apps1'),
                get_condition('', 'equals', True, 'DeploymentGroup')
            )},
            {'DeployRevision': Equals(Ref('UpdateMode'), 'CodeDeploy')},
        ])

        # Resources
        # FOR SINGLEAPP CODEDEPLOY
        if len(cfg.Apps) == 1:
            R_DeploymentGroup = CDDeploymentGroup('DeploymentGroup')
            R_DeploymentGroup.setup(index=1)

            add_obj(R_DeploymentGroup)


class CD_Applications(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            App = cdd.Application(key + n)
            App.ApplicationName = get_endvalue(key + n)

            add_obj(App)



import troposphere.certificatemanager as crm

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class CRMCertificate(crm.Certificate):
    def setup(self, key):
        auto_get_props(self, key, recurse=True)


# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class CRM_Certificate(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            if 'Enabled' in v and not v['Enabled']:
                continue
            # resources
            resname = key + n
            r_Certificate = CRMCertificate(resname)
            r_Certificate.setup(key=v)
            r_Certificate.Tags = Tags(Name=n)

            add_obj([
                r_Certificate,
            ])

            # outputs
            o_Certificate = Output(resname)
            o_Certificate.Value = Ref(resname)

            if hasattr(r_Certificate, 'Condition'):
                o_Certificate.Condition = r_Certificate.Condition

            add_obj([
                o_Certificate,
            ])


class CRM_CertificateEC2(object):
    def __init__(self, key):
        if cfg.ListenerLoadBalancerHttpsPort != 'None':
            CRM_Certificate(key)


class CRM_CertificateECS(object):
    def __init__(self, key):
        if cfg.ListenerLoadBalancerHttpsPort != 443:
            CRM_Certificate(key)


class CRM_CertificateRES(object):
    def __init__(self, key):
        CRM_Certificate(key)



#!/usr/bin/env python3
import sys
import os
import json

from troposphere import Template
from awsibox import args, cfg, RP, discover, generate

args = args.get_args()


def output_template(template, brand=None, role=None):
    if args.Output == 'json':
        output = template.to_json()
        extension = '.json'
    elif args.Output == 'cjson':
        output = template.to_json(indent=None)
        extension = '.json'
    elif args.Output == 'yaml':
        output = template.to_yaml()
        extension = '.yaml'

    if args.action == 'view':
        print(output)
    elif args.action == 'write':
        file_path = os.path.join(os.getcwd(), 'templates', brand)
        try:
            os.makedirs(file_path)
        except:
            pass
        file_name = os.path.join(file_path, role + extension) 
        with open(file_name, 'w') as f:
            f.write(output + '\n')


def get_template():
    # save base cfg 
    cfg_base = cfg.__dict__.copy()

    cfg.template = Template()
    RP.build_RP()
    template = generate.generate()

    # restore base cfg
    cfg.__dict__.clear()
    cfg.__dict__.update(cfg_base)

    return template


if args.action == 'view':
    cfg.envrole = args.EnvRole
    cfg.brand = args.Brand

    template = get_template()
    
    output_template(template)

elif args.action == 'write':
    discover_map = discover.discover(args.Brands, args.EnvRoles, args.StackTypes)
    
    for brand, roles in discover_map.items():
        cfg.brand = brand
    
        for role in roles:
            cfg.envrole = role

            print('Brand: %s - EnvRole: %s' % (brand, role))

            try:
                template = get_template()
            except Exception as e:
                print('ERROR: %s' % e)
                exit(1)

            output_template(template, brand, role)



from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj, auto_build_obj)


class CFM_Parameters(object):
    def __init__(self, key):
        auto_build_obj(Parameter(''), getattr(cfg, key))


class CFM_Conditions(object):
    def __init__(self, key):
        do_no_override(True)
        for n, v in getattr(cfg, key).items():
            c_Condition = {n: eval(v)}

            add_obj(c_Condition)
        do_no_override(False)


class CFM_Mappings(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            c_Mapping = {n: v}

            cfg.Mappings.update(c_Mapping)


class CFM_Outputs(object):
    def __init__(self, key):
        auto_build_obj(Output(''), getattr(cfg, key))



import troposphere.elasticloadbalancing as elb
import troposphere.elasticloadbalancingv2 as elbv2
import troposphere.ec2 as ec2

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .route53 import R53_RecordSetEC2LoadBalancer, R53_RecordSetECSLoadBalancer
from .securitygroup import (SecurityGroupRuleELBPorts,SecurityGroupIngressInstanceELBPorts,
    SecurityGroupLoadBalancer, SecurityGroupsIngressEcs)
from .lambdas import LambdaPermissionLoadBalancing


# Temporary fix for https://github.com/cloudtools/troposphere/issues/1474
elbv2.one_of = my_one_of

# S - CLASSIC LOAD BALANCING #
class ELBListener(elb.Listener):
    def setup(self, key):
        auto_get_props(self, key)
        self.PolicyNames = [
            If('LoadBalancerCookieSticky', 'LBCookieStickinessPolicy', Ref('AWS::NoValue'))
        ]
        self.SSLCertificateId = If(
            'ListenerLoadBalancerHttpsPort',
            If(
                'LoadBalancerSslCertificateAdHoc',
                Ref('CertificateLoadBalancerAdHocExternal'),
                get_endvalue('RegionalCertificateArn')
            ),
            Ref('AWS::NoValue')
        )


class ELBLoadBalancer(elb.LoadBalancer):
    def setup(self):
        self.AccessLoggingPolicy = If(
            'LoadBalancerLog',
            elb.AccessLoggingPolicy(
                EmitInterval=get_endvalue('LoadBalancerLog'),
                Enabled=True,
                S3BucketName=Sub(cfg.BucketLogs),
                S3BucketPrefix=''
            ),
            Ref('AWS::NoValue')
        )
        self.ConnectionDrainingPolicy = elb.ConnectionDrainingPolicy(
            Enabled=True,
            Timeout=5
        )
        self.ConnectionSettings = elb.ConnectionSettings(
            IdleTimeout=get_endvalue('LoadBalancerIdleTimeout')
        )
        self.CrossZone = True
        self.HealthCheck = elb.HealthCheck(
            HealthyThreshold=get_endvalue('HealthyThresholdCount'),
            Interval=get_endvalue('HealthCheckIntervalSeconds'),
            Target=get_endvalue('HealthCheckTarget'),
            Timeout=get_endvalue('HealthCheckTimeoutSeconds'),
            UnhealthyThreshold=get_endvalue('UnhealthyThresholdCount')
        )
        self.LBCookieStickinessPolicy = If(
            'LoadBalancerCookieSticky',
            [
                elb.LBCookieStickinessPolicy(
                    PolicyName='LBCookieStickinessPolicy',
                    CookieExpirationPeriod=get_endvalue('LoadBalancerCookieSticky')
                )
            ],
            Ref('AWS::NoValue')
        )
        self.SecurityGroups = [GetAtt('SecurityGroupLoadBalancer', 'GroupId')]


class ELBLoadBalancerExternal(ELBLoadBalancer):
    def setup(self):
        super(ELBLoadBalancerExternal, self).setup()
        self.Scheme = 'internet-facing'
        self.Subnets = Split(',', get_expvalue('SubnetsPublic'))


class ELBLoadBalancerInternal(ELBLoadBalancer):
    def setup(self):
        super(ELBLoadBalancerInternal, self).setup()
        self.Scheme = 'internal'
        self.Subnets = Split(',', get_expvalue('SubnetsPrivate'))
# E - CLASSIC LOAD BALANCING #


# S - V2 LOAD BALANCING #
class ELBV2Listener(elbv2.Listener):
    def setup(self, scheme):  # Ex. scheme = External/Internal
        self.DefaultActions = [elbv2.Action(
            Type='forward',
            TargetGroupArn=Ref('TargetGroup' + scheme)
        )]
        self.LoadBalancerArn = Ref('LoadBalancerApplication' + scheme)


class ELBV2ListenerHttp(ELBV2Listener):
    def setup(self, **kwargs):
        super(ELBV2ListenerHttp, self).setup(**kwargs)
        self.Port = get_endvalue('ListenerLoadBalancerHttpPort')
        self.Protocol = 'HTTP'


class ELBV2ListenerHttps(ELBV2Listener):
    def setup(self, **kwargs):
        super(ELBV2ListenerHttps, self).setup(**kwargs)
        self.Certificates = [elbv2.Certificate(
            CertificateArn=If(
                'LoadBalancerSslCertificateAdHoc',
                Ref('CertificateLoadBalancerAdHocExternal'),
                get_endvalue('RegionalCertificateArn')
            )
        )]
        self.Port = get_endvalue('ListenerLoadBalancerHttpsPort')
        self.Protocol = 'HTTPS'


class ELBV2TargetGroup(elbv2.TargetGroup):
    def setup(self):
        self.HealthCheckIntervalSeconds = get_endvalue('HealthCheckIntervalSeconds')
        self.HealthCheckTimeoutSeconds = get_endvalue('HealthCheckTimeoutSeconds')
        if cfg.HealthCheckPath != 'None':
            self.HealthCheckPath = get_endvalue('HealthCheckPath')
        self.HealthyThresholdCount = get_endvalue('HealthyThresholdCount')
        self.UnhealthyThresholdCount = get_endvalue('UnhealthyThresholdCount')
        self.TargetGroupAttributes = [
            elbv2.TargetGroupAttribute(
                Key='deregistration_delay.timeout_seconds',
                Value=get_endvalue('TargetGroupDeregistrationDelay')
            ),
            elbv2.TargetGroupAttribute(
                Key='stickiness.enabled',
                Value=If(
                    'TargetGroupCookieSticky',
                    'true',
                    'false',
                )
            ),
            If(
                'TargetGroupCookieSticky',
                elbv2.TargetGroupAttribute(
                    Key='stickiness.lb_cookie.duration_seconds',
                    Value=get_endvalue('TargetGroupCookieSticky')
                ),
                Ref('AWS::NoValue')
            )
        ]
        self.VpcId = get_expvalue('VpcId')


class ELBV2TargetGroupEC2(ELBV2TargetGroup):
    def setup(self):
        super(ELBV2TargetGroupEC2, self).setup()
        self.Port = get_endvalue('Listeners1InstancePort')
        self.Protocol = get_endvalue('Listeners1Protocol')


class ELBV2TargetGroupALB(elbv2.TargetGroup):
    def setup(self, lambda_arn):
        self.Targets = [
            elbv2.TargetDescription(
                Id=lambda_arn,
            )
        ]
        self.TargetType = 'lambda'


class ELBV2TargetGroupECS(ELBV2TargetGroup):
    def setup(self):
        super(ELBV2TargetGroupECS, self).setup()
        self.Port = get_endvalue('ContainerDefinitions1ContainerPort')
        self.Protocol = get_endvalue('ContainerDefinitions1Protocol')
        self.TargetType = If(
            'NetworkModeAwsVpc',
            'ip',
            Ref('AWS::NoValue')
        )


class ELBV2ListernerRuleECS(elbv2.ListenerRule):
    def setup(self, key, index, mapname, scheme):
        auto_get_props(self, key, mapname='ListenerRules%s' % index, recurse=True)
        if 'Conditions' not in key:
            self.Conditions = []
            if 'HostHeader' in key:
                self.Conditions.append(elbv2.Condition(
                    Field='host-header',
                    Values=[get_endvalue(mapname + 'HostHeader', issub=True)]
                ))
            if 'PathPattern' in key:
                self.Conditions.append(elbv2.Condition(
                    Field='path-pattern',
                    Values=[Sub('${Value}', **{'Value': get_endvalue(mapname + 'PathPattern')})]
                ))
        if 'Actions' not in key:
            self.Actions = [elbv2.Action(
                Type='forward',
                TargetGroupArn=Ref('TargetGroup' + scheme)
            )]
        self.Priority = get_endvalue(mapname + 'Priority')


class ELBV2LoadBalancer(elbv2.LoadBalancer):
    def setup(self):
        self.LoadBalancerAttributes = [
            If(
                'LoadBalancerLog',
                {
                    'Key': 'access_logs.s3.bucket',
                    'Value': Sub(cfg.BucketLogs)
                },
                Ref('AWS::NoValue')
            ),
            If(
                'LoadBalancerLog',
                {
                    'Key': 'access_logs.s3.enabled',
                    'Value': True
                },
                {
                    'Key': 'access_logs.s3.enabled',
                    'Value': False
                },
            ),
            If(
                'LoadBalancerLog',
                {
                    'Key': 'access_logs.s3.prefix',
                    'Value': Sub('${EnvRole}.${AWS::StackName}')
                },
                Ref('AWS::NoValue')
            ),
            If(
                'LoadBalancerHttp2',
                {
                    'Key': 'routing.http2.enabled',
                    'Value': get_endvalue('LoadBalancerHttp2')
                },
                Ref('AWS::NoValue')
            ),
        ]
        self.SecurityGroups = [
            GetAtt('SecurityGroupLoadBalancer', 'GroupId')
        ]


class ELBV2LoadBalancerExternal(ELBV2LoadBalancer):
    def setup(self):
        super(ELBV2LoadBalancerExternal, self).setup()
        self.Scheme = 'internet-facing'
        self.Subnets = Split(',', get_expvalue('SubnetsPublic'))


class ELBV2LoadBalancerInternal(ELBV2LoadBalancer):
    def setup(self):
        super(ELBV2LoadBalancerInternal, self).setup()
        self.Scheme = 'internal'
        self.Subnets = Split(',', get_expvalue('SubnetsPrivate'))


class ELBV2LoadBalancerExternalALB(ELBV2LoadBalancerExternal):
    def setup(self):
        super(ELBV2LoadBalancerExternalALB, self).setup()
        self.Condition = 'LoadBalancerApplicationExternal'
        self.SecurityGroups = [
            get_expvalue('SecurityGroupLoadBalancerApplicationExternal')
        ]

class ELBV2LoadBalancerInternalALB(ELBV2LoadBalancerInternal):
    def setup(self):
        super(ELBV2LoadBalancerInternalALB, self).setup()
        self.Condition = 'LoadBalancerApplicationInternal'
        self.SecurityGroups = [
            get_expvalue('SecurityGroupLoadBalancerApplicationInternal')
        ]


class ELBV2ListenerAction404(elbv2.Action):
    def __init__(self, **kwargs):
        super(ELBV2ListenerAction404, self).__init__(**kwargs)
        self.FixedResponseConfig =elbv2.FixedResponseConfig('')
        self.FixedResponseConfig.ContentType = 'text/plain'
        self.FixedResponseConfig.MessageBody = '404 Not Found\n'
        self.FixedResponseConfig.StatusCode = '404'
        self.Type = 'fixed-response'


# E - V2 LOAD BALANCING #


# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class LB_Listeners(object):
    def __init__(self):
        # Conditions
        for i in cfg.AllowedIp:
            condname = 'AllowedIp%s' % i  # Ex. AllowedIp1
            add_obj(get_condition(condname, 'not_equals', 'None', condname + 'Enabled'))


class LB_ListenersEC2(LB_Listeners):
    def __init__(self):
        super(LB_ListenersEC2, self).__init__()
        # Resources
        Listeners = []
        Securitygroup_Rules = []
        SecuritygroupIngress_InstanceRules = []
        for n, v in cfg.Listeners.items():
            mapname = 'Listeners%s' % n  # Ex Listeners1

            if cfg.LoadBalancerClassic:
                Listener = ELBListener(mapname)
                Listener.setup(key=v)
                Listeners.append(Listener)

            for i in cfg.AllowedIp:
                ipname = 'AllowedIp%s' % i   # Ex. AllowedIp1
                condnameprivate = 'SecurityGroupRulePrivate' + mapname + ipname  # Ex. SecurityGroupRulePrivateListeners1AllowedIp1
                condnamepublic = 'SecurityGroupRulePublic' + mapname # Ex. SecurityGroupRulePublicListeners1

                # conditions
                add_obj(
                    {condnameprivate: And(
                        Condition(ipname),
                        get_condition('', 'equals', 'Private', mapname + 'LoadBalancerAccess')
                    )}
                )

                SGRule = SecurityGroupRuleELBPorts(mapname)
                SGRule.setup()
                SGRule.CidrIp = get_endvalue(ipname + 'Ip')

                Securitygroup_Rules.append(
                    If(
                        condnameprivate,
                        SGRule,
                        Ref('AWS::NoValue'),
                    )
                )

            # conditions
            add_obj(get_condition(condnamepublic, 'equals', 'Public', mapname + 'LoadBalancerAccess'))

            SGRule = SecurityGroupRuleELBPorts(mapname)
            SGRule.setup()
            SGRule.CidrIp = '0.0.0.0/0'

            Securitygroup_Rules.append(
                If(
                    condnamepublic,
                    SGRule,
                    Ref('AWS::NoValue'),
                )
            )

            # resources
            r_SGIInstance = SecurityGroupIngressInstanceELBPorts('SecurityGroupIngress' + mapname)
            r_SGIInstance.setup(listener=mapname)

            add_obj(r_SGIInstance)

            # outputs
            Listener_Output = Output(mapname)
            Listener_Output.Value = Sub(
                '${LoadBalancerPort}.${Protocol}.${LoadBalancerAccess}',
                **{
                    'LoadBalancerPort': get_endvalue(mapname + 'LoadBalancerPort'),
                    'Protocol': get_endvalue(mapname + 'Protocol'),
                    'LoadBalancerAccess': get_endvalue(mapname + 'LoadBalancerAccess')
                }
            )
            add_obj(Listener_Output)

        R_SG = SecurityGroupLoadBalancer('SecurityGroupLoadBalancer')
        R_SG.setup()
        R_SG.SecurityGroupIngress = Securitygroup_Rules

        add_obj([
            R_SG,
        ])

        # Outputs
        O_SG = Output('SecurityGroupLoadBalancer')
        O_SG.Value = Ref('SecurityGroupLoadBalancer')

        add_obj([
            O_SG,
        ])

        self.Listeners = Listeners


class LB_ListenerRulesExternalInternal(object):
    def __init__(self, index, key, mapname, scheme):
        # resources
        R_RuleHttp = ELBV2ListernerRuleECS('ListenerHttp' + scheme + 'Rules' + index)  # Ex. ListenerHttpExternalRules1
        R_RuleHttp.setup(key=key, index=index, mapname=mapname, scheme=scheme)
        R_RuleHttp.ListenerArn=get_expvalue('ListenerHttpDefault' + scheme, 'LoadBalancerApplicationStack')
        R_RuleHttp.Condition = 'ListenerLoadBalancerHttpPort'

        R_RuleHttps = ELBV2ListernerRuleECS('ListenerHttps' + scheme + 'Rules' + index)
        R_RuleHttps.setup(key=key, index=index, mapname=mapname, scheme=scheme)
        R_RuleHttps.ListenerArn=get_expvalue('ListenerHttpsDefault' + scheme, 'LoadBalancerApplicationStack')
        R_RuleHttps.Condition = 'ListenerLoadBalancerHttpsPort'

        # Create ListenerRule only in stack's specific new Listener
        ListenerHttpPort = cfg.ListenerLoadBalancerHttpPort
        ListenerHttpsPort = cfg.ListenerLoadBalancerHttpsPort
        Protocol = key['Protocol'] if 'Protocol' in key else 'auto'
        RuleHttpAdd = None
        RuleHttpsAdd = None

        if 'NoDefault' in key:
            R_RuleHttp.ListenerArn = Ref('ListenerHttp' + scheme)
            R_RuleHttps.ListenerArn = Ref('ListenerHttps' + scheme)
            if ListenerHttpPort != 80:
                RuleHttpAdd = True
            if ListenerHttpsPort != 443:
                RuleHttpsAdd = True
        else:
            # by default create http rules on Internal LB and https rules on External one
            if scheme == 'Internal':
                RuleHttpAdd = True
            if scheme == 'External':
                RuleHttpsAdd = True

        # can be forced or overriden by key http/https/any
        if Protocol == 'http':
            RuleHttpAdd = True
            RuleHttpsAdd = None
        if Protocol == 'https':
            RuleHttpAdd = None
            RuleHttpsAdd = True
        if Protocol == 'any':
            RuleHttpAdd = True
            RuleHttpsAdd = True

        if RuleHttpAdd:
            add_obj(R_RuleHttp)
        if RuleHttpsAdd:
            add_obj(R_RuleHttps)


class LB_ListenerRules(object):
    def __init__(self):
        for n, v in cfg.ListenerRules.items():
            mapname = 'ListenerRules%s' % n  # Ex. ListenerRules1

            # parameters
            p_Priority = Parameter(mapname + 'Priority')
            p_Priority.Description = 'Listener Rule Priority, lesser value = high priority - empty for default based on env/role'

            add_obj(p_Priority)

            ListenerRule_Out_String = ['Priority=${Priority}']
            ListenerRule_Out_Map = {'Priority': get_endvalue(mapname + 'Priority')}

            if 'HostHeader' in v:
                p_HostHeader = Parameter(mapname + 'HostHeader')
                p_HostHeader.Description = 'Listener Rule HostHeader Condition - empty for default based on env/role'

                add_obj(p_HostHeader)

                # outputs
                ListenerRule_Out_String.append('HostHeader=${HostHeader}')
                ListenerRule_Out_Map.update({
                    'HostHeader': get_endvalue(mapname + 'HostHeader', issub=True)
                })

            if 'PathPattern' in v:
                p_PathPattern = Parameter(mapname + 'PathPattern')
                p_PathPattern.Description = 'Listener Rule PathPattern Condition - empty for default based on env/role'

                add_obj(p_PathPattern)

                # outputs
                ListenerRule_Out_String.append('PathPattern=${PathPattern}')
                ListenerRule_Out_Map.update({
                    'PathPattern': get_endvalue(mapname + 'PathPattern', issub=True)
                })

            # resources
            if cfg.LoadBalancerApplicationExternal:
                LB_ListenerRulesExternalInternal(index=str(n), key=v, mapname=mapname, scheme='External')

            if cfg.LoadBalancerApplicationInternal:
                LB_ListenerRulesExternalInternal(index=str(n), key=v, mapname=mapname, scheme='Internal')

            # outputs
            o_ListenerRule = Output(mapname)
            o_ListenerRule.Value = Sub(','.join(ListenerRule_Out_String), **ListenerRule_Out_Map)

            add_obj(o_ListenerRule)


class LB_ListenersV2ExternalInternal(object):
    def __init__(self, scheme):
        # Conditions
        add_obj([
            {'SecurityGroupIngressPublicLoadBalancerHttp' + scheme: And(
                Condition('ListenerLoadBalancerHttpPort'),
                Condition('LoadBalancerPublic')
            )},
            {'SecurityGroupIngressPublicLoadBalancerHttps' + scheme: And(
                Condition('ListenerLoadBalancerHttpsPort'),
                Condition('LoadBalancerPublic')
            )}
        ])

        # Resources
        R_SGIHttpPublic = SecurityGroupsIngressEcs('SecurityGroupIngressPublicLoadBalancerHttp' + scheme)
        R_SGIHttpPublic.setup(proto='Http', scheme=scheme)
        R_SGIHttpPublic.CidrIp = '0.0.0.0/0'

        R_SGIHttpsPublic = SecurityGroupsIngressEcs('SecurityGroupIngressPublicLoadBalancerHttps' + scheme)
        R_SGIHttpsPublic.setup(proto='Https', scheme=scheme)
        R_SGIHttpsPublic.CidrIp = '0.0.0.0/0'

        add_obj([
            R_SGIHttpPublic,
            R_SGIHttpsPublic,
        ])

        for i in cfg.AllowedIp:
            ipname = 'AllowedIp%s' % i  # Ex. AllowedIp1
            condnamehttpprivate = 'SecurityGroupIngressPrivateLoadBalancerHttp' + scheme + ipname
            condnamehttpsprivate = 'SecurityGroupIngressPrivateLoadBalancerHttps' + scheme + ipname
            
            # conditions
            add_obj([
                {condnamehttpprivate: And(
                    Condition(ipname),
                    Not(Condition('LoadBalancerPublic')),
                    Condition('ListenerLoadBalancerHttpPort')
                )},
                {condnamehttpsprivate: And(
                    Condition(ipname),
                    Not(Condition('LoadBalancerPublic')),
                    Condition('ListenerLoadBalancerHttpsPort')
                )}
            ])

            # resources
            SGIHttpPrivate = SecurityGroupsIngressEcs(condnamehttpprivate)
            SGIHttpPrivate.setup(proto='Http', scheme=scheme)
            SGIHttpPrivate.CidrIp = get_endvalue(ipname + 'Ip')

            SGIHttpsPrivate = SecurityGroupsIngressEcs(condnamehttpsprivate)
            SGIHttpsPrivate.setup(proto='Https', scheme=scheme)
            SGIHttpsPrivate.CidrIp = get_endvalue(ipname + 'Ip')

            add_obj([
                SGIHttpPrivate,
                SGIHttpsPrivate,
            ])

        if cfg.ListenerLoadBalancerHttpPort != 80:
            R_ListenerHttp = ELBV2ListenerHttp('ListenerHttp' + scheme)  # Ex. ListenerHttpExternal
            R_ListenerHttp.setup(scheme=scheme)
            R_ListenerHttp.LoadBalancerArn = get_expvalue('LoadBalancerApplication' + scheme, 'LoadBalancerApplicationStack')

            add_obj(R_ListenerHttp)

        if cfg.ListenerLoadBalancerHttpsPort != 443 and scheme == 'External':
            R_ListenerHttps = ELBV2ListenerHttps('ListenerHttps' + scheme)
            R_ListenerHttps.setup(scheme=scheme)
            R_ListenerHttps.Condition = 'ListenerLoadBalancerHttpsPort'
            R_ListenerHttps.LoadBalancerArn = get_expvalue('LoadBalancerApplication' + scheme, 'LoadBalancerApplicationStack')

            add_obj(R_ListenerHttps)


class LB_ListenersV2EC2(LB_ListenersEC2):
    def __init__(self):
        super(LB_ListenersV2EC2, self).__init__()
        # Resources
        if cfg.LoadBalancerApplicationExternal:
            R_ListenerHttpExternal = ELBV2ListenerHttp('ListenerHttpExternal')
            R_ListenerHttpExternal.setup(scheme='External')
            R_ListenerHttpExternal.Condition = 'ListenerLoadBalancerHttpPort'

            R_ListenerHttpsExternal = ELBV2ListenerHttps('ListenerHttpsExternal')
            R_ListenerHttpsExternal.setup(scheme='External')
            R_ListenerHttpsExternal.Condition = 'ListenerLoadBalancerHttpsPort'
       
            add_obj([
                R_ListenerHttpExternal,
                R_ListenerHttpsExternal,
            ])

        if cfg.LoadBalancerApplicationInternal:
            R_ListenerHttpInternal = ELBV2ListenerHttp('ListenerHttpInternal')
            R_ListenerHttpInternal.setup(scheme='Internal')
            R_ListenerHttpInternal.Condition = 'ListenerLoadBalancerHttpPort'

            add_obj([
                R_ListenerHttpInternal,
            ])
        
        LB_TargetGroupsEC2()
 

class LB_ListenersV2ECS(LB_Listeners):
    def __init__(self):
        super(LB_ListenersV2ECS, self).__init__()
        # Resources
        if cfg.LoadBalancerApplicationExternal:
            LB_ListenersV2ExternalInternal(scheme='External')

        if cfg.LoadBalancerApplicationInternal:
            LB_ListenersV2ExternalInternal(scheme='Internal')

        LB_TargetGroupsECS()
        LB_ListenerRules()


class LB_ListenersV2ALB(object):
    def __init__(self):
        # Resources
        if cfg.LoadBalancerApplicationExternal:
            # CREATE SPECIFIC CLASS - EXTERNAL INTERNAL AS PARAMETER
            R_ListenerHttp = ELBV2ListenerHttp('ListenerHttpDefaultExternal')
            R_ListenerHttp.setup(scheme='External')
            # Now that AWS allow to send a fixed response, there is no need to have a DefaultTarget Group pointing to nothing
            # R_ListenerHttp.DefaultActions[0].TargetGroupArn=Ref('TargetGroupDefaultExternal')
            R_ListenerHttp.DefaultActions[0] = ELBV2ListenerAction404()
            R_ListenerHttp.Condition = 'LoadBalancerApplicationExternal'

            R_ListenerHttps = ELBV2ListenerHttps('ListenerHttpsDefaultExternal')
            R_ListenerHttps.setup(scheme='External')
            # Now that AWS allow to send a fixed response, there is no need to have a DefaultTarget Group pointing to nothing
            # R_ListenerHttps.DefaultActions[0].TargetGroupArn=Ref('TargetGroupDefaultExternal')
            R_ListenerHttps.DefaultActions[0] = ELBV2ListenerAction404()
            R_ListenerHttps.Certificates[0].CertificateArn = get_endvalue('RegionalCertificateArn')
            R_ListenerHttps.Condition = 'LoadBalancerApplicationExternal'

            add_obj([
                R_ListenerHttp,
                R_ListenerHttps,
            ])

            # Outputs
            O_ListenerHttp = Output('ListenerHttpDefaultExternal')
            O_ListenerHttp.Condition = 'LoadBalancerApplicationExternal'
            O_ListenerHttp.Value = Ref('ListenerHttpDefaultExternal')
            O_ListenerHttp.Export = Export(Sub('ListenerHttpDefaultExternal-${AWS::StackName}'))

            O_ListenerHttps = Output('ListenerHttpsDefaultExternal')
            O_ListenerHttps.Condition = 'LoadBalancerApplicationExternal'
            O_ListenerHttps.Value = Ref('ListenerHttpsDefaultExternal')
            O_ListenerHttps.Export = Export(Sub('ListenerHttpsDefaultExternal-${AWS::StackName}'))

            add_obj([
                O_ListenerHttp,
                O_ListenerHttps,
            ])

        if cfg.LoadBalancerApplicationInternal:
            R_ListenerHttp = ELBV2ListenerHttp('ListenerHttpDefaultInternal')
            R_ListenerHttp.setup(scheme='Internal')
            # Now that AWS allow to send a fixed response, there is no need to have a DefaultTarget Group pointing to nothing
            # R_ListenerHttp.DefaultActions[0].TargetGroupArn=Ref('TargetGroupDefaultInternal')
            R_ListenerHttp.DefaultActions[0] = ELBV2ListenerAction404()
            R_ListenerHttp.Condition = 'LoadBalancerApplicationInternal'

            add_obj([
                R_ListenerHttp,
            ])            
            
            # Outputs
            O_ListenerHttp = Output('ListenerHttpDefaultInternal')
            O_ListenerHttp.Condition = 'LoadBalancerApplicationInternal'
            O_ListenerHttp.Value = Ref('ListenerHttpDefaultInternal')
            O_ListenerHttp.Export = Export(Sub('ListenerHttpDefaultInternal-${AWS::StackName}'))

            add_obj([
                O_ListenerHttp,
            ])


class LB_TargetGroups(object):
    def __init__(self):
        # Conditions
        add_obj(get_condition('TargetGroupCookieSticky', 'not_equals', 'None'))


class LB_TargetGroupsEC2(LB_TargetGroups):
    def __init__(self):
        super(LB_TargetGroupsEC2, self).__init__()
        # Resources
        if cfg.LoadBalancerApplicationExternal:
            R_TargetGroup = ELBV2TargetGroupEC2('TargetGroupExternal')
            R_TargetGroup.setup()

            add_obj(R_TargetGroup)

            cfg.Alarm['TargetEC2External5XX']['Enabled'] = True

        if cfg.LoadBalancerApplicationInternal:
            R_TargetGroup = ELBV2TargetGroupEC2('TargetGroupInternal')
            R_TargetGroup.setup()

            add_obj(R_TargetGroup)
            
            cfg.Alarm['TargetEC2Internal5XX']['Enabled'] = True


class LB_TargetGroupsECS(LB_TargetGroups):
    def __init__(self):
        super(LB_TargetGroupsECS, self).__init__()
        # Resources
        if cfg.LoadBalancerApplicationExternal:
            R_TargetGroup = ELBV2TargetGroupECS('TargetGroupExternal')
            R_TargetGroup.setup()

            add_obj(R_TargetGroup)

            cfg.Alarm['TargetExternal5XX']['Enabled'] = True

        if cfg.LoadBalancerApplicationInternal:
            R_TargetGroup = ELBV2TargetGroupECS('TargetGroupInternal')
            R_TargetGroup.setup()

            add_obj(R_TargetGroup)
            
            cfg.Alarm['TargetInternal5XX']['Enabled'] = True


class LB_TargetGroupsALB(object):
    def __init__(self):
        lambda_name = 'ServiceUnavailable'
        lambda_arn = get_expvalue('Lambda' + lambda_name + 'Arn')
        perm_name = 'LambdaPermission' + lambda_name + 'LoadBalancerApplication'

        # Resources
        if cfg.LoadBalancerApplicationExternal:
            R_TargetGroup = ELBV2TargetGroupALB('TargetGroupServiceUnavailableExternal')
            R_TargetGroup.Condition = 'LoadBalancerApplicationExternal'
            R_TargetGroup.DependsOn = perm_name + 'External'
            R_TargetGroup.setup(lambda_arn=lambda_arn)

            R_Permission = LambdaPermissionLoadBalancing(perm_name + 'External')
            R_Permission.Condition = 'LoadBalancerApplicationExternal'
            R_Permission.setup(name=lambda_arn)

            O_TargetGroup = Output('TargetGroupServiceUnavailableExternal')
            O_TargetGroup.Condition = 'LoadBalancerApplicationExternal'
            O_TargetGroup.Value = Ref('TargetGroupServiceUnavailableExternal')

            add_obj([
                R_TargetGroup,
                R_Permission,
                O_TargetGroup,
            ])

        if cfg.LoadBalancerApplicationInternal:
            R_TargetGroup = ELBV2TargetGroupALB('TargetGroupServiceUnavailableInternal')
            R_TargetGroup.Condition = 'LoadBalancerApplicationInternal'
            R_TargetGroup.DependsOn = perm_name + 'Internal'
            R_TargetGroup.setup(lambda_arn=lambda_arn)
            
            R_Permission = LambdaPermissionLoadBalancing(perm_name + 'Internal')
            R_Permission.Condition = 'LoadBalancerApplicationInternal'
            R_Permission.setup(name=lambda_arn)

            O_TargetGroup = Output('TargetGroupServiceUnavailableInternal')
            O_TargetGroup.Condition = 'LoadBalancerApplicationInternal'
            O_TargetGroup.Value = Ref('TargetGroupServiceUnavailableInternal')

            add_obj([
                R_TargetGroup,
                R_Permission,
                O_TargetGroup,
            ])
            

class LB_ElasticLoadBalancingApplication(object):
    def __init__(self):
        # Parameters
        P_Http2 = Parameter('LoadBalancerHttp2')
        P_Http2.Description = 'Load Balancer Http2 - empty for default based on env/role'
        P_Http2.AllowedValues = ['', 'true', 'false']

        add_obj([
            P_Http2,
        ])

        # Conditions
        add_obj(get_condition('LoadBalancerHttp2', 'not_equals', 'None'))


class LB_ElasticLoadBalancingClassicEC2(LB_ListenersEC2):
    def __init__(self):
        super(LB_ElasticLoadBalancingClassicEC2, self).__init__()
        # Conditions
        add_obj(get_condition('LoadBalancerCookieSticky', 'not_equals', 'None'))

        # Resources
        if cfg.LoadBalancerClassicExternal:
            R_ELBExternal = ELBLoadBalancerExternal('LoadBalancerClassicExternal')
            R_ELBExternal.setup()
            R_ELBExternal.Listeners = self.Listeners

            add_obj(R_ELBExternal)
            
            cfg.Alarm['BackendExternal5XX']['Enabled'] = True

        if cfg.LoadBalancerClassicInternal:
            R_ELBInternal = ELBLoadBalancerInternal('LoadBalancerClassicInternal')
            R_ELBInternal.setup()
            R_ELBInternal.Listeners = self.Listeners

            add_obj(R_ELBInternal)
            
            cfg.Alarm['BackendInternal5XX']['Enabled'] = True

        # Outputs
        O_HealthCheck = Output('HealthCheck')
        O_HealthCheck.Value = Sub(
            'Type=${Type},Target=${Target},Interval=${Interval},Timeout=${Timeout},Healthy=${Healthy},Unhealthy=${Unhealthy}',
            **{
                'Type': get_endvalue('HealthCheckType'),
                'Target': get_endvalue('HealthCheckTarget') if cfg.HealthCheckTarget != 'None' else '',
                'Interval': get_endvalue('HealthCheckIntervalSeconds'),
                'Timeout': get_endvalue('HealthCheckTimeoutSeconds'),
                'Healthy': get_endvalue('HealthyThresholdCount'),
                'Unhealthy': get_endvalue('UnhealthyThresholdCount'),
            }
        )

        add_obj([
            O_HealthCheck,
        ])


class LB_ElasticLoadBalancingApplicationEC2(object):
    def __init__(self):
        LB_ElasticLoadBalancingApplication()
        # Resources
        if cfg.LoadBalancerApplicationExternal:
            R_ELBExternal = ELBV2LoadBalancerExternal('LoadBalancerApplicationExternal')
            R_ELBExternal.setup()

            add_obj([
                R_ELBExternal,
            ])
       
        if cfg.LoadBalancerApplicationInternal:
            R_ELBInternal = ELBV2LoadBalancerInternal('LoadBalancerApplicationInternal')
            R_ELBInternal.setup()

            add_obj([
                R_ELBInternal,
            ])
        
        LB_ListenersV2EC2()
            
        # outputs
        O_HealthCheck = Output('HealthCheck')
        O_HealthCheck.Value = Sub(
            'Type=${Type},Path=${Path},Interval=${Interval},Timeout=${Timeout},Healthy=${Healthy},Unhealthy=${Unhealthy}',
            **{
                'Type': get_endvalue('HealthCheckType'),
                'Path': get_endvalue('HealthCheckPath') if cfg.HealthCheckPath != 'None' else '',
                'Interval': get_endvalue('HealthCheckIntervalSeconds'),
                'Timeout': get_endvalue('HealthCheckTimeoutSeconds'),
                'Healthy': get_endvalue('HealthyThresholdCount'),
                'Unhealthy': get_endvalue('UnhealthyThresholdCount'),
            }
        )
            
        add_obj([
            O_HealthCheck,
        ])



class LB_ElasticLoadBalancingEC2(object):
    def __init__(self, key):
        # Resources
        R53_RecordSetEC2LoadBalancer()
        
        if cfg.LoadBalancerClassic:
            LB_ElasticLoadBalancingClassicEC2()

        if cfg.LoadBalancerApplication:
            LB_ElasticLoadBalancingApplicationEC2()


class LB_ElasticLoadBalancingALB(object):
    def __init__(self, key):
        LB_ElasticLoadBalancingApplication()

        if cfg.LoadBalancerApplicationExternal:
            # Resources
            R_ELB = ELBV2LoadBalancerExternalALB('LoadBalancerApplicationExternal')
            R_ELB.setup()

            add_obj(R_ELB)

            # Outputs
            O_ELB = Output('LoadBalancerApplicationExternal')
            O_ELB.Condition = 'LoadBalancerApplicationExternal'
            O_ELB.Value = Ref('LoadBalancerApplicationExternal')
            O_ELB.Export = Export(Sub('LoadBalancerApplicationExternal-${AWS::StackName}'))

            O_ELBDNS = Output('LoadBalancerApplicationExternalDNS')
            O_ELBDNS.Condition = 'LoadBalancerApplicationExternal'
            O_ELBDNS.Value = GetAtt('LoadBalancerApplicationExternal', 'DNSName')
            O_ELBDNS.Export = Export(Sub('LoadBalancerApplicationExternalDNS-${AWS::StackName}'))

            O_ELBFullName = Output('LoadBalancerApplicationExternalFullName')
            O_ELBFullName.Condition = 'LoadBalancerApplicationExternal'
            O_ELBFullName.Value = GetAtt('LoadBalancerApplicationExternal', 'LoadBalancerFullName')
            O_ELBFullName.Export = Export(Sub('LoadBalancerApplicationExternalFullName-${AWS::StackName}'))

            add_obj([
                O_ELB,
                O_ELBDNS,
                O_ELBFullName,
            ])

        if cfg.LoadBalancerApplicationInternal:
            # Resources
            R_ELB = ELBV2LoadBalancerInternalALB('LoadBalancerApplicationInternal')
            R_ELB.setup()

            add_obj(R_ELB)

            # Outputs
            O_ELB = Output('LoadBalancerApplicationInternal')
            O_ELB.Condition = 'LoadBalancerApplicationInternal'
            O_ELB.Value = Ref('LoadBalancerApplicationInternal')
            O_ELB.Export = Export(Sub('LoadBalancerApplicationInternal-${AWS::StackName}'))

            O_ELBDNS = Output('LoadBalancerApplicationInternalDNS')
            O_ELBDNS.Condition = 'LoadBalancerApplicationInternal'
            O_ELBDNS.Value = GetAtt('LoadBalancerApplicationInternal', 'DNSName')
            O_ELBDNS.Export = Export(Sub('LoadBalancerApplicationInternalDNS-${AWS::StackName}'))

            O_ELBFullName = Output('LoadBalancerApplicationInternalFullName')
            O_ELBFullName.Condition = 'LoadBalancerApplicationInternal'
            O_ELBFullName.Value = GetAtt('LoadBalancerApplicationInternal', 'LoadBalancerFullName')
            O_ELBFullName.Export = Export(Sub('LoadBalancerApplicationInternalFullName-${AWS::StackName}'))

            add_obj([
                O_ELB,
                O_ELBDNS,
                O_ELBFullName,
            ])

        # Resources
        LB_ListenersV2ALB()
        # Create TargetGroups pointing to LambdaServiceUnavailable
        try: cfg.ServiceUnavailable
        except:
            pass
        else:
            LB_TargetGroupsALB()


class LB_ElasticLoadBalancingECS(object):
    def __init__(self, key):
        # Resources
        LB_ListenersV2ECS()
        R53_RecordSetECSLoadBalancer()



from pprint import pprint, pformat
import argparse

from . import cfg


# parse main argumets
def get_args():
    parser = argparse.ArgumentParser(description='Generate AWS CloudFormation Stack Templates')

    # subparser
    subparsers = parser.add_subparsers(help='Desired Action', dest='action')

    # parent parser args
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('-R', '--Regions', help='Regions to enable', nargs='+', default=cfg.DEFAULT_REGIONS)
    parent_parser.add_argument('-o', '--Output', help='Output Format', choices=['json', 'cjson', 'yaml'], default='json')

    # view parser
    parser_view = subparsers.add_parser('view', parents=[parent_parser], help='Generate template and print to standard output')
    parser_view.add_argument('-r', '--EnvRole', help='Stack EnvRole', required=True, type=str)
    parser_view.add_argument('-b', '--Brand', help='Brand', required=True, type=str)
    parser_view.add_argument('-d', '--Debug', help='Show RP Dict', action='store_true')

    # write parser
    parser_write = subparsers.add_parser('write', parents=[parent_parser], help='Generate template and write to "template/${brand}/$envrole" file')

    parser_write.add_argument('-b', '--Brands', help='Limit generation to specified Brands', nargs='+')

    roletype_group_write = parser_write.add_mutually_exclusive_group(required=False)
    roletype_group_write.add_argument('-r', '--EnvRoles', help='Limit generation to specified EnvRoles', nargs='+')
    roletype_group_write.add_argument('-t', '--StackTypes', help='Limit generation to specified StackTypes', nargs='+', default=[''])

    args = parser.parse_args()

    try:
        cfg.debug = args.Debug
    except:
        cfg.debug = None
    cfg.regions = args.Regions

    return args



import sys
import os
import mmap

from . import cfg


def get_brands():
    brand_int = os.listdir(cfg.PATH_INT)
    brand_ext = os.listdir(cfg.PATH_EXT)

    brands = set(brand_int + brand_ext)

    brands.remove('BASE')

    return brands


def get_files(brand):
    files = []

    path_int = os.path.join(cfg.PATH_INT, brand)
    path_ext = os.path.join(cfg.PATH_EXT, brand)

    for n in [path_int, path_ext]:
        for root, directories, filenames in os.walk(n, topdown=True):
            try:
                directories.remove('UNUSED')
            except:
                pass
            for filename in filenames:
                if not filename.startswith('.'):
                    files.append(os.path.join(root,filename))

    return files


def build_discover_map(brand, files, stacktypes):
    roles = []

    for n in files:
        try:
            with open(n, 'r', encoding='utf-8') as f:
                s = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)

                if any(s.find((pattern + t).encode('utf-8')) != -1 for t in stacktypes):
                    base_name = os.path.basename(n)
                    role = os.path.splitext(base_name)[0]
                    roles.append(role)
        except IOError:
            pass

    if brand == 'BASE':
        for n in ext_brands:
            add_to_map(n, roles)
    else:
        add_to_map(brand, roles)


def add_to_map(brand, roles):
    global discover_map

    try:
        discover_map[brand].extend(roles)
    except:
        discover_map[brand] = roles


def discover(brands, envroles, stacktypes):
    global discover_map
    global pattern
    global ext_brands

    pattern = ' StackType: '

    discover_map = {}

    if brands:
        ext_brands = list(brands)
    else:
        ext_brands = get_brands()
        brands = list(ext_brands)

    brands.append('BASE')

    for brand in brands:
        if envroles:
            files = []
            for role in envroles:
                files.append(os.path.join(cfg.PATH_EXT, brand, role + '.yml'))
                if brand == 'BASE':
                    files.append(os.path.join(cfg.PATH_INT, brand, role + '.yml'))
        else:
            files = get_files(brand)

        build_discover_map(brand, files, stacktypes)

    return discover_map



import os
from collections import OrderedDict, Mapping
from functools import reduce

no_override = False

Parameters = OrderedDict()
Conditions = OrderedDict()
Mappings = OrderedDict()
Resources = OrderedDict()
Outputs = OrderedDict()

PATH_INT = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'cfg')
PATH_INT = os.path.normpath(PATH_INT)

PATH_EXT = os.path.join(os.getcwd(), 'cfg')
PATH_EXT = os.path.normpath(PATH_EXT)

MAX_SECURITY_GROUPS = 4
# SECURITY_GROUPS_DEFAULT equals list of empty values (Ex. for "MAX_SECURITY_GROUPS = 3" we have "SECURITY_GROUPS_DEFAULT = ',,'")
SECURITY_GROUPS_DEFAULT = reduce(lambda a, i: ',' + str(a), list(range(MAX_SECURITY_GROUPS - 1)), '')

ENV_BASE = ['dev', 'stg', 'prd']

DEFAULT_REGIONS = [
    'eu-west-1',
    'us-east-1',
    'eu-central-1',
]

AZones = {
    'MAX': 6,
    'default': 3,
    'us-east-1': 6,
    'us-west-1': 2,
    'us-west-2': 4,
    'ca-central-1': 2,
    'sa-east-1': 2,
    'cn-north-1': 2,
    'ap-northeast-3': 1,
}

AZoneNames = ['A', 'B', 'C', 'D', 'E', 'F']

PARAMETERS_SKIP_OVERRIDE_CONDITION = (
    'Env',
    'UpdateMode',
    'RecordSetExternal',
    'DoNotSignal',
    'EfsMounts',
    'ImageIdLatest',
    'VPCCidrBlock',
    'VPCName',
)

EVAL_FUNCTIONS_IN_CFG = (
    'cfg.',
    'get_expvalue(',
    'Sub(',
    'Ref(',
    'get_subvalue(',
    'get_endvalue(',
    'get_condition(',
    'GetAtt(',
    'Split(',
    'Export(',
    'Join(',
    'If(',
    'dict(',
    'eval(',
    'Tags(',
    'str(',
)

BASE_CFGS = {
    'Bucket': {
        'Enabled': 'None',
        'Create': 'None',
        'AccountsRO': OrderedDict([
            ('dev', 'None'),
            ('stg', 'None'),
            ('prd', 'None'),
        ]),
        'Cors': 'None',
        'CloudFrontOriginAccessIdentityExtra': OrderedDict([
            ('Dev', 'None'),
            ('Stg', 'None'),
            ('Prd', 'None'),
        ]),
        'ReplicaDstRegion': 'None',
        'ReplicaDstOwner': 'None',
        'ReplicaSrcAccount': 'None',
        'Versioning': 'None',
    }
}

INSTANCE_LIST = [
    'default',
    'm3.medium', 'm3.large', 'm3.xlarge', 'm3.2xlarge',
    'c3.large', 'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge',
    'r3.large', 'r3.xlarge', 'r3.2xlarge', 'r3.4xlarge',
    'i2.xlarge', 'i2.2xlarge', 'i2.4xlarge',
    'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge',
    't2.nano', 't2.micro', 't2.small', 't2.medium', 't2.large', 't2.xlarge', 't2.2xlarge',
    'm4.large', 'm4.xlarge', 'm4.2xlarge', 'm4.4xlarge',
    'c4.large', 'c4.xlarge', 'c4.2xlarge', 'c4.4xlarge',
    'r4.large', 'r4.xlarge', 'r4.2xlarge', 'r4.4xlarge',
    'g2.2xlarge', 'g3s.xlarge',
    't3.nano', 't3.micro', 't3.small', 't3.medium', 't3.large', 't3.xlarge', 't3.2xlarge',
    'm5.large', 'm5.xlarge', 'm5.2xlarge', 'm5.4xlarge', 'm5.8xlarge', 'm5.12xlarge', 'm5.16xlarge', 'm5.24xlarge',
    'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge', 'c5.9xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge',
    'r5.large', 'r5.xlarge', 'r5.2xlarge', 'r5.4xlarge', 'r5.8xlarge', 'r5.12xlarge', 'r5.16xlarge', 'r5.24xlarge',
    'a1.medium', 'a1.large', 'a1.xlarge', 'a1.2xlarge', 'a1.4xlarge',
    'm5a.large', 'm5a.xlarge', 'm5a.2xlarge', 'm5a.4xlarge', 'm5a.8xlarge', 'm5a.12xlarge', 'm5a.16xlarge', 'm5a.24xlarge',
    'r5a.large', 'r5a.xlarge', 'r5a.2xlarge', 'r5a.4xlarge', 'r5a.8xlarge', 'r5a.12xlarge', 'r5a.16xlarge', 'r5a.24xlarge',
]

CFG_TO_CLASS = OrderedDict([
    ('MappingClass', {'module': 'mappings', 'class': 'Mappings'}),
    ('Parameter', {'module': 'cloudformation', 'class': 'CFM_Parameters'}),
    ('Condition', {'module': 'cloudformation', 'class': 'CFM_Conditions'}),
    ('Mapping', {'module': 'cloudformation', 'class': 'CFM_Mappings'}),
    ('CapacityDesired', {'module': 'autoscaling', 'class': 'AS_Autoscaling'}),
    ('ScalingPolicyUpScalingAdjustment1', {'module': 'autoscaling', 'class': 'AS_ScalingPoliciesStep'}),
    ('ScalingPolicyTrackings', {'module': 'autoscaling', 'class': 'AS_ScalingPoliciesTracking'}),
    ('Bucket', {'module': 's3', 'class': 'S3_Buckets'}),
    ('BucketPolicy', {'module': 's3', 'class': 'S3_BucketPolicies'}),
    ('Certificate', {'module': 'crm', 'class': 'CRM_Certificate'}),
    ('CodeDeployApp', {'module': 'codedeploy', 'class': 'CD_Applications'}),
    ('Repository', {'module': 'ecr', 'class': 'ECR_Repositories'}),
    ('Cluster', {'module': 'ecs', 'class': 'ECS_Cluster'}),
    ('Service', {'module': 'ecs', 'class': 'ECS_Service'}),
    ('ApiGatewayAccount', {'module': 'apigateway', 'class': 'AGW_Account'}),
    ('ApiGatewayDomainName', {'module': 'apigateway', 'class': 'AGW_DomainName'}),
    ('ApiGatewayResource', {'module': 'apigateway', 'class': 'AGW_RestApi'}),
    ('ApiGatewayStage', {'module': 'apigateway', 'class': 'AGW_Stages'}),
    ('ApiGatewayUsagePlan', {'module': 'apigateway', 'class': 'AGW_UsagePlans'}),
    ('ApiGatewayApiKey', {'module': 'apigateway', 'class': 'AGW_ApiKeys'}),
    ('LogGroupName', {'module': 'logs', 'class': 'LGS_LogGroup'}),
    ('ContainerDefinitions', {'module': 'ecs', 'class': 'ECS_TaskDefinition'}),
    ('EFSFileSystem', {'module': 'efs', 'class': 'EFS_FileStorage'}),
    ('CacheSubnetGroup', {'module': 'elasticache', 'class': 'CCH_SubnetGroups'}),
    ('CacheNodeType', {'module': 'elasticache', 'class': 'CCH_Cache'}),
    ('EventsRule', {'module': 'events', 'class': 'EVE_EventRules'}),
    ('Role', {'module': 'iam', 'class': 'IAM_Roles'}),
    ('IAMPolicy', {'module': 'iam', 'class': 'IAM_Policies'}),
    ('IAMUser', {'module': 'iam', 'class': 'IAM_Users'}),
    ('IAMGroup', {'module': 'iam', 'class': 'IAM_Groups'}),
    ('IAMUserToGroupAddition', {'module': 'iam', 'class': 'IAM_UserToGroupAdditions'}),
    ('KMSKey', {'module': 'kms', 'class': 'KMS_Keys'}),
    ('Lambda', {'module': 'lambdas', 'class': 'LBD_Lambdas'}),
    ('LambdaLayerVersion', {'module': 'lambdas', 'class': 'LBD_LayerVersions'}),
    ('LambdaPermission', {'module': 'lambdas', 'class': 'LBD_Permissions'}),
    ('LambdaEventSourceMapping', {'module': 'lambdas', 'class': 'LBD_EventSourceMappings'}),
    ('LoadBalancerApplication', {'module': 'loadbalancing', 'class': 'LB_ElasticLoadBalancing'}),
    ('LoadBalancerClassic', {'module': 'loadbalancing', 'class': 'LB_ElasticLoadBalancing'}),
    ('Alarm', {'module': 'cloudwatch', 'class': 'CW_Alarms'}),
    ('CloudFront', {'module': 'cloudfront', 'class': 'CF_CloudFront'}),
    ('SNSSubscription', {'module': 'sns', 'class': 'SNS_Subscriptions'}),
    ('SNSTopic', {'module': 'sns', 'class': 'SNS_Topics'}),
    ('SQSQueue', {'module': 'sqs', 'class': 'SQS_Queues'}),
    ('ASGLifecycleHook', {'module': 'autoscaling', 'class': 'AS_LifecycleHook'}),
    ('DBInstanceClass', {'module': 'rds', 'class': 'RDS_DB'}),
    ('DBParameterGroup', {'module': 'rds', 'class': 'RDS_ParameterGroups'}),
    ('DBSubnetGroup', {'module': 'rds', 'class': 'RDS_SubnetGroups'}),
    ('HostedZone', {'module': 'route53', 'class': 'R53_HostedZones'}),
    ('WafByteMatchSet', {'module': 'waf', 'class': ['WAF_GlobalByteMatchSets', 'WAF_RegionalByteMatchSets']}),
    ('WafIPSet', {'module': 'waf', 'class': ['WAF_GlobalIPSets', 'WAF_RegionalIPSets']}),
    ('WafRule', {'module': 'waf', 'class': ['WAF_GlobalRules', 'WAF_RegionalRules']}),
    ('WafWebAcl', {'module': 'waf', 'class': ['WAF_GlobalWebAcls', 'WAF_RegionalWebAcls']}),
    ('ServiceDiscovery', {'module': 'servicediscovery', 'class': 'SRVD_ServiceDiscovery'}),
    ('VPCEndpoint', {'module': 'vpc', 'class': 'VPC_Endpoint'}),
    ('SecurityGroupBase', {'module': 'securitygroup', 'class': 'SG_SecurityGroup'}),
    ('SecurityGroupIngress', {'module': 'securitygroup', 'class': 'SG_SecurityGroupIngressesExtra'}),
    ('VPC', {'module': 'vpc', 'class': 'VPC_VPC'}),
    # Output need to be last line
    ('Output', {'module': 'cloudformation', 'class': 'CFM_Outputs'}),
])



import troposphere.ec2 as ec2

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class EC2VPCPeeringConnection(ec2.VPCPeeringConnection):
    def setup(self):
        self.Condition = 'VPCPeeringConnection'
        self.PeerVpcId = get_expvalue('VpcId-stg', '')
        self.VpcId = get_expvalue('VpcId-dev', '')
        self.Tags = Tags(Name='Dev-Peer-Staging')


class EC2RouteNatGateway(ec2.Route):
    def setup(self):
        self.DestinationCidrBlock = '0.0.0.0/0'
        self.NatGatewayId = Ref('NatGateway')
        self.RouteTableId = Ref('RouteTablePrivate')


class EC2RouteInternetGateway(ec2.Route):
    def setup(self):
        self.DestinationCidrBlock = '0.0.0.0/0'
        self.GatewayId = Ref('InternetGateway')
        self.RouteTableId = Ref('RouteTablePublic')


class EC2RoutePeeringConnection(ec2.Route):
    def setup(self):
        self.Condition = 'VPCPeeringConnection'
        self.VpcPeeringConnectionId = Ref('VPCPeeringConnectionDevStaging')


class EC2RoutePeeringConnectionDev(EC2RoutePeeringConnection):
    def setup(self):
        super(EC2RoutePeeringConnectionDev, self).setup()
        self.DestinationCidrBlock = get_expvalue('VPCCidr-stg', '')
        self.RouteTableId = get_expvalue('RouteTablePrivate-dev', '')


class EC2RoutePeeringConnectionStg(EC2RoutePeeringConnection):
    def setup(self):
        super(EC2RoutePeeringConnectionStg, self).setup()
        self.DestinationCidrBlock = get_expvalue('VPCCidr-dev', '')
        self.RouteTableId = get_expvalue('RouteTablePrivate-stg', '')


class EC2VPCEndpoint(ec2.VPCEndpoint):
    def setup(self):
        self.RouteTableIds = [ get_expvalue('RouteTablePrivate') ]
        self.VpcId = get_expvalue('VpcId')


class EC2VPCEndpointS3(EC2VPCEndpoint):
    def setup(self):
        super(EC2VPCEndpointS3, self).setup()
        self.ServiceName = Sub('com.amazonaws.${AWS::Region}.s3')

class EC2VPC(ec2.VPC):
    def setup(self):
        self.CidrBlock = Ref('VPCCidrBlock')
        self.EnableDnsSupport = True
        self.EnableDnsHostnames = True


class EC2Subnet(ec2.Subnet):
    def setup(self, zone):
        self.AvailabilityZone = Sub('${AWS::Region}%s' % zone.lower()) 
        self.VpcId = Ref('VPC')


class EC2SubnetPrivate(EC2Subnet):
    def setup(self, zone):
        super(EC2SubnetPrivate, self).setup(zone)
        self.CidrBlock = Ref('SubnetCidrBlockPrivate%s' % zone)
        self.MapPublicIpOnLaunch = False
        self.Tags = Tags(Name=Sub('${VPCName}-Private%s' % zone))


class EC2SubnetPublic(EC2Subnet):
    def setup(self, zone):
        super(EC2SubnetPublic, self).setup(zone)
        self.CidrBlock = Ref('SubnetCidrBlockPublic%s' % zone)
        self.MapPublicIpOnLaunch = True
        self.Tags = Tags(Name=Sub('${VPCName}-Public%s' % zone))


class EC2RouteTable(ec2.RouteTable):
    def setup(self):
        self.VpcId = Ref('VPC')




class EC2SubnetRouteTableAssociationPrivate(ec2.SubnetRouteTableAssociation):
    def setup(self, zone):
        self.RouteTableId = Ref('RouteTablePrivate')
        self.SubnetId = Ref('SubnetPrivate%s' % zone)


class EC2SubnetRouteTableAssociationPublic(ec2.SubnetRouteTableAssociation):
    def setup(self, zone):
        self.RouteTableId = Ref('RouteTablePublic')
        self.SubnetId = Ref('SubnetPublic%s' % zone)
##

class VPC_Endpoint(object):
    def __init__(self, key):
        # Conditions
        add_obj(get_condition('EC2VPCEndpointS3', 'not_equals', 'None', 'VPCEndpoint'))

        # Resources
        R_S3 = EC2VPCEndpointS3('EC2VPCEndpointS3')
        R_S3.setup()
        R_S3.Condition = 'EC2VPCEndpointS3'

        add_obj([
            R_S3,
        ])


class VPC_VPC(object):
    def __init__(self, key):
        vpc_net = '10.80'
        o_subnetprivate = []
        o_subnetpublic = []

        # Resources
        R_VPC = ec2.VPC('VPC')
        auto_get_props(R_VPC, mapname='')

        R_RouteTablePrivate = EC2RouteTable('RouteTablePrivate')
        R_RouteTablePrivate.setup()
        R_RouteTablePrivate.Tags = Tags(Name=Sub('${VPCName}-Private'))

        R_RouteTablePublic = EC2RouteTable('RouteTablePublic')
        R_RouteTablePublic.setup()
        R_RouteTablePublic.Tags = Tags(Name=Sub('${VPCName}-Public'))

        R_InternetGateway = ec2.InternetGateway('InternetGateway')
        R_InternetGateway.Tags = Tags(Name=Ref('VPCName'))

        R_VPCGatewayAttachment = ec2.VPCGatewayAttachment('VPCGatewayAttachment')
        R_VPCGatewayAttachment.InternetGatewayId = Ref('InternetGateway')
        R_VPCGatewayAttachment.VpcId = Ref('VPC')

        R_EIPNat = ec2.EIP('EIPNat')
        R_EIPNat.Domain = 'vpc'

        R_NatGateway = ec2.NatGateway('NatGateway')
        R_NatGateway.AllocationId = GetAtt('EIPNat', 'AllocationId')
        R_NatGateway.SubnetId = Ref('SubnetPublicA')

        R_RouteNatGateway = EC2RouteNatGateway('RouteNatGateway')
        R_RouteNatGateway.setup()

        R_RouteInternetGateway = EC2RouteInternetGateway('RouteInternetGateway')
        R_RouteInternetGateway.setup()

        add_obj([
            R_VPC,
            R_RouteTablePrivate,
            R_RouteTablePublic,
            R_InternetGateway,
            R_VPCGatewayAttachment,
            R_EIPNat,
            R_NatGateway,
            R_RouteNatGateway,
            R_RouteInternetGateway,
        ])

        for i in range(cfg.AZones['MAX']):
            zone_name = cfg.AZoneNames[i]
            zone_cond = 'Zone%s' % zone_name

            # parameters
            p_SubnetCidrBlockPrivate = Parameter('SubnetCidrBlockPrivate%s' % zone_name )
            p_SubnetCidrBlockPrivate.Description = 'Ip Class Range for Private Subnet in Zone %s' % zone_name
            p_SubnetCidrBlockPrivate.Default = '%s.%s.0/20' %(vpc_net, i * 16)

            p_SubnetCidrBlockPublic = Parameter('SubnetCidrBlockPublic%s' % zone_name )
            p_SubnetCidrBlockPublic.Description = 'Ip Class Range for Public Subnet in zone %s' % zone_name
            p_SubnetCidrBlockPublic.Default = '%s.%s.0/24' %(vpc_net, i + 1)

            add_obj([
                p_SubnetCidrBlockPrivate,
                p_SubnetCidrBlockPublic,
            ])

            # conditions
            add_obj(
                {zone_cond: Equals(
                    FindInMap('AvabilityZones', Ref('AWS::Region'), 'Zone%s' % i), 'True'
                )},
            )

            # resources

            r_SubnetPrivate = EC2SubnetPrivate('SubnetPrivate%s' % zone_name)
            r_SubnetPrivate.setup(zone_name)
            r_SubnetPrivate.Condition = zone_cond

            r_SubnetPublic = EC2SubnetPublic('SubnetPublic%s' % zone_name)
            r_SubnetPublic.setup(zone_name)
            r_SubnetPublic.Condition = zone_cond

            r_SubnetRouteTableAssociationPrivate = EC2SubnetRouteTableAssociationPrivate('SubnetRouteTableAssociationPrivate%s' % zone_name)
            r_SubnetRouteTableAssociationPrivate.setup(zone_name)
            r_SubnetRouteTableAssociationPrivate.Condition = zone_cond

            r_SubnetRouteTableAssociationPublic = EC2SubnetRouteTableAssociationPublic('SubnetRouteTableAssociationPublic%s' % zone_name)
            r_SubnetRouteTableAssociationPublic.setup(zone_name)
            r_SubnetRouteTableAssociationPublic.Condition = zone_cond

            add_obj([
                r_SubnetPrivate,
                r_SubnetPublic,
                r_SubnetRouteTableAssociationPrivate,
                r_SubnetRouteTableAssociationPublic,
            ])

            # outputs
            o_subnetprivate.append(If(
                zone_cond,
                Ref('SubnetPrivate%s' % zone_name),
                Ref('AWS::NoValue')
            ))

            o_subnetpublic.append(If(
                zone_cond,
                Ref('SubnetPublic%s' % zone_name),
                Ref('AWS::NoValue')
            ))


        # Outputs
        O_SubnetsPrivate = Output('SubnetsPrivate')
        O_SubnetsPrivate.Value = Join(',', o_subnetprivate)
        O_SubnetsPrivate.Export = Export('SubnetsPrivate')

        O_SubnetsPublic = Output('SubnetsPublic')
        O_SubnetsPublic.Value = Join(',', o_subnetpublic)
        O_SubnetsPublic.Export = Export('SubnetsPublic')

        add_obj([
            O_SubnetsPrivate,
            O_SubnetsPublic,
        ])
